/****************************************************************************
	Entry system for tksesh according to the Manuel de Codage's specifications
										(c) 1993-2001 Serge Rosmorduc																	 
								Manuel de Codage to Tcl List Builder.
													designed for lemon

 lemon is a parser generator " la yacc", but it does handle memory in a better
 and safer way.
****************************************************************************/


%include{

#define _POSIX_SOURCE 1

#include <stddef.h>							/* NULL, size_t, and so on */
#include <malloc.h>
#include <tcl.h>
#include "general.h"
#include "TclListUtils.h"

extern int 
seshlex();

void 
yyerror(char *s);

#include <stdio.h>
#ifndef NULL
#define NULL 0
#endif

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define ParseTrace  Mdc2lst_ParseTrace
#define ParseTokenName  Mdc2lst_ParseTokenName
#define ParseAlloc  Mdc2lst_ParseAlloc
#define ParseFree  Mdc2lst_ParseFree
#define Parse  Mdc2lst_Parse

 
 struct MdcParseContext {
	 Tcl_Interp *interp;
	 Tcl_Obj *result;
 };
}

%extra_argument {struct MdcParseContext *context}

%token_type {Tcl_Obj *}

%type optWordEnd int
%type optGrammar int

// Note :

// TopLevelToken : a token which can only appear outside a construct, be it a cartouche
// or a cadrat (e.g. ENDOFLINE)
// StandAloneToken : a token which can't appear inside a cadrat, but is legal in a cartouche
// BasicToken : the rest of the tokens.

// mdcfile   ::= [WORDEND] [SEPARATOR] (textitem [SEPARATOR])*

mdcfile(A) ::= startingGarbage textitems(B). {A= B;}
mdcfile(A) ::= error. {A= NULL;}

startingGarbage ::= WORDEND.
startingGarbage ::= .

textitems(A) ::= . {A= Tcl_NewListObj(0,0);}
textitems(A) ::= textitems(B) optSeparator textitem(C). {
	Tcl_ListObjAppendElement(context->interp, B, C);
	A= B;
}

// textitem    ::= topcadrat | cartouche | PAGEEND | LINEEND | TEXTSUPER | TOGGLE | TEXT

textitem(A) ::= topcadrat(B).
{
	A= B;
}

textitem(A) ::= cartouche(B). {
	A= B;
}

textitem(A) ::= LINEEND(B). {
	A= B;
}

textitem(A) ::= PAGEEND(B). {
	A= B;
}

textitem(A) ::= TEXTSUPER(B). {
	A= B;
}

textitem(A) ::= TOGGLE(B). {
	A= B;
}

textitem(A) ::= TEXT(B). {
	A= B;
}

// topcadrat ::= cadrat [SHADING]

topcadrat(A) ::= cadrat(B) optShading(C). {
	/* Format for cadrat : {c DATA ?opt1 val1? ?opt2 val2?} */
	/* so we simply append shading : */
	Tcl_ListObjAppendList(context->interp, B, C);
	A= B;
}
 
optShading(A)::= . {	A= Tcl_NewListObj(0,0);
}

optShading(A)::= SHADING(B). {
	/*
	 * Shading is a integer :
	 * 1 means first quarter is shaded
	 * 2 means second quarter is shaded
	 * 4 means third quarter is shaded
	 * 8 means fourth quarter is shaded
	 *
	 * combining these is done by adding the corresponding values.
	 * 
	 * Thus, if both first, second and fourth quarters are shaded,
	 * shading's value will be 1+2+8= 11
	 */
	A= Tksesh_NewTclCouple(Tksesh_NewTclKeyword("shading"), B);
}

// cadrat ::= subcadrat (":" subcadrat)*

cadrat(A) ::= subcadrats(B). {
	A= Tksesh_NewTclCouple(Tksesh_NewTclKeyword("c"),B);
}

subcadrats(A) ::= subcadrat(B). {
	A= Tksesh_NewTclSingleton(B);
}

subcadrats(A) ::= subcadrats(B) COLUMN subcadrat(C).
{
	Tcl_ListObjAppendElement(context->interp, B, C);
	A= B;
}

// subcadrat  ::= inhierlist ("*" inhierlist)*

subcadrat(A) ::= inhierlists(B).
{
	A= B;
}

inhierlists(A) ::= inhierlist(B).
{
	A= Tksesh_NewTclSingleton(B);
}

inhierlists(A) ::= inhierlists(B) STAR inhierlist(C).
{
	Tcl_ListObjAppendElement(context->interp, B, C);
	A= B;
}

// inhierlist  ::= complexsign | BEGINPHIL cadrats ENDPHIL | "(" cadrats ")"

inhierlist(A) ::= complexsign(B).
{
	A= B;
}

inhierlist(A) ::= BEGINPHIL(B) cadrats(C) ENDPHIL(D).
{	
		A= Tksesh_NewTclCouple(Tksesh_NewTclKeyword("philology"),C);
		A= B;
		A= D;
	/* Now, compute and add information about what kind of philological
     comments we have : TO DO */
	
}

inhierlist(A) ::= BPAR cadrats(B) EPAR.
{
	A= B;
}

// cadrats   ::= maincadrat ([SEPARATOR] maincadrat)*

cadrats(A) ::= maincadrats(B).
{
	A= Tksesh_NewTclCouple(Tksesh_NewTclKeyword("subcadrat"),B);
}

maincadrats(A) ::= maincadrat(B).
{
	A= Tksesh_NewTclSingleton(B);
}

maincadrats(A) ::= maincadrats(B) optSeparator maincadrat(C).
{
	Tcl_ListObjAppendElement(context->interp, B, C);
	A= B;
}

optSeparator ::= .
optSeparator ::= SEPARATOR.

// maincadrat ::= cadrat | TEXT

maincadrat(A) ::= cadrat(B).
{
	A= B;
}

maincadrat(A) ::= TEXT(B).
{
	A= Tksesh_NewTclCouple(Tksesh_NewTclKeyword("latin"),B);
}

// complexsign ::= (hieroglyphs | SYMBOL) [OVERWRITE (hieroglyphs|SYMBOL)]

complexsign(A) ::= hieroglyphsOrSymbol(B).
{
	A= B;
}

complexsign(A) ::= hieroglyphsOrSymbol(B) OVERWRITE hieroglyphsOrSymbol(C).
{
	A= Tksesh_NewTclTriplet(Tksesh_NewTclKeyword("overwrite"), B, C);
}

hieroglyphsOrSymbol(A) ::= hieroglyphs(B).
{
	A= B;
}

hieroglyphsOrSymbol(A) ::= SYMBOL(B).
{
	A= B;
}

// hieroglyphs ::= hieroglyph [ligature]
// ligature ::=  ("&" hieroglyph)+

hieroglyphs(A) ::= ligatures(B).
{
	int l;
	Tcl_ListObjLength(context->interp, B, &l);
	if (l == 1) {
		A= B;
	} else {
		A= Tksesh_NewTclCouple(Tksesh_NewTclKeyword("lig"), B);
	}
}

ligatures(A) ::= hieroglyph(B).
{
	A= Tksesh_NewTclSingleton(B);
}

ligatures(A) ::= hieroglyphs(B) AMP hieroglyph(C).
{
	Tcl_ListObjAppendElement(context->interp, B, C);
	A= B;
}

// hieroglyph  ::= [GRAMMAR] HIEROGLYPH MODIFIER* [WORDEND]

hieroglyph(A) ::= optGrammar(B) HIEROGLYPH(C) modifiers(D) optWordEnd(E).
{
		A= Tksesh_NewTclCouple(Tksesh_NewTclKeyword("s"), C);
		Tcl_ListObjAppendList(context->interp, A, D);
		if (E) {
				Tcl_ListObjAppendElement(context->interp, A, 
																 Tksesh_NewTclCouple(Tksesh_NewTclKeyword("wordend"),
																										 Tcl_NewIntObj(1)));
		}
		if (B) {
				Tcl_ListObjAppendElement(context->interp, A, 
																 Tksesh_NewTclCouple(Tksesh_NewTclKeyword("gram"),
																										 Tcl_NewIntObj(1)));
		}
}

optWordEnd(A) ::= .
{
	A= FALSE;
}

optWordEnd(A) ::= WORDEND.
{
	A= TRUE;
}

optGrammar(A) ::= .
{
	A= FALSE;
}

optGrammar(A) ::= GRAMMAR.
{
	A= TRUE;
}

modifiers(A) ::=.
{
	A= Tcl_NewListObj(0,0);
}

modifiers(A) ::= modifiers(B) MODIFIER(C).
{
	Tcl_ListObjAppendElement(context->interp, B, C);
	A= B;
}

// The new kind of cartouches.
// cartouche ::= BEGINCARTOUCHE (cadrat [HACHING] | TEXT | TOGGLE)* ENDCARTOUCHE

cartouche(A) ::= BEGINCARTOUCHE cartoucheTexts(C) ENDCARTOUCHE.
{
	A= Tksesh_NewTclCouple(Tksesh_NewTclKeyword("cartouche"), C);
}

cartouche(A) ::= BEGINOLDCARTOUCHE cartoucheTexts(C) ENDCARTOUCHE.
{
	A= Tksesh_NewTclCouple(Tksesh_NewTclKeyword("cartouche"), C);
}

cartoucheTexts(A) ::= .
{
	A= Tcl_NewListObj(0,0);
}

cartoucheTexts(A) ::= cartoucheTexts(B) topcadrat(C).
{
	Tcl_ListObjAppendElement(context->interp, B, C);
	A= B;
}

cartoucheTexts(A) ::= cartoucheTexts(B) TEXT(C).
{
	Tcl_ListObjAppendElement(context->interp, B, C);
	A= B;
}

cartoucheTexts(A) ::= cartoucheTexts(B) TOGGLE(C).
{
	Tcl_ListObjAppendElement(context->interp, B, C);
	A= B;
}

%code {
	
	
	/**
	 */
	
	void
		parseMDC(char * mdc) {
		
		pParser = ParseAlloc( malloc );
		
		while( GetNextToken(pTokenizer, &hTokenId, &sToken) ){
			/* Parse( parser, token type, token value, context) */
			Parse(pParser, hTokenId, sToken, &sState);
		}
		/* notify the end of file */
		Parse(pParser, 0, sToken, &sState);
		
		/* free the parser */
		ParseFree(pParser, free );
		
	}
}


/*
 * Local Variables:
 * header: "seshaux.h"
 * tab-width: 2
 * End:
 */
