/* seshdb.c --
	 interface between the database functions in btdb.[ch]
	 and Tcl
*/

/*  1998,2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: seshdb.c,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 */

/*
 * $Log: seshdb.c,v $
 * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * Imported using tkCVS
 *
 * Revision 1.12  2000/07/22 12:25:04  rosmord
 * Deletion function added. Some test dones
 *
 * Revision 1.11  1999/10/04 16:32:30  rosmord
 * CW compatible
 *
 */


#define _POSIX_SOURCE 1

/*
	 interface between Tcl and btdb

	 USE :

	 basic functions :
	 set db [open_db filename]
	 set txt [get_db $db $key]
	 put_db $db $key $value
	 close_db $db

	 open_db also accepts arguments "r" for read_only, write for read/write (default)
	 and a block size for the data. The block size must be between 10 and 255

	 advanced functions : indexes. They reference a place in the base.

	 set i [first_db_idx $db]

	 or : set i [get_db_idx $db $key] (index to the place with the first element
	 greater than or equal to $key

	 incr_db_idx $i $val : moves in the base by $val record. Returns 1 if move is possible, else 0.
	 
	 getfrom_db_idx [-key] $i : sends back a list with two values
	 (one only with -key option), corresponding to the key and value of
	 the element pointed by $i.

	 free_db_idx $i : frees index $i

	 islast_db_idx $i
*/

#include <ctype.h>
#include <tcl.h>
#include <fcntl.h>							/* fd manipulation */
#include <string.h>
#include <stdlib.h>

#include "general.h"
#include "btdb.h"
#include "chaine.h"

#define MAXFILES 256

#include "fordll.h"

/*
 * Declarations for functions defined in this file.
 */

EXTERN EXPORT(int,Dbinter_Init) _ANSI_ARGS_((Tcl_Interp *interp));

struct dbf_struct {
	int used;
	btdb_t file;
};							/* up to 256 files ? */

struct idx_struct {
	int used;
	btdb_index_t idx;
};

struct local_data {
	Chaine files;									/* opened files */
	Chaine indexes;								/* indexes to files */
};


#define DBF(n) (*((struct dbf_struct *)get_from_Chaine(data->files,(n))))
#define NODBF(n) ((n >= size_of_Chaine(data->files)) || ! DBF(n).used)

#define IDX(n) (*((struct idx_struct *)get_from_Chaine(data->indexes,(n))))
#define NOIDX(n) ((n >= size_of_Chaine(data->indexes)) || ! IDX(n).used)

static struct local_data *
new_local_data()
{
	struct local_data *result;
	result= (struct local_data *)ALLOUER(sizeof(struct local_data));
	init_Chaine(&(result->files), sizeof(struct dbf_struct));
	init_Chaine(&(result->indexes), sizeof(struct idx_struct));
	return result;
}

static void 
delete_local_data(ClientData clientData,
									Tcl_Interp *interp)
{
	int i;
	struct local_data *data;
	data= (struct local_data *) clientData;
	
	for (i= 0; i < size_of_Chaine(data->files); i++)
		if (DBF(i).used)
			btdb_close(&(DBF(i).file));
	free_Chaine(data->files); 
	free_Chaine(data->indexes);
	LIBERE(data);
}

static int
get_new_index(ClientData clientData)
{
	int num;
	struct local_data *data;
	data= (struct local_data *) clientData;
	num= 0;
	while(num < size_of_Chaine(data->indexes) && IDX(num).used)
		num++;
	if (num == size_of_Chaine(data->indexes))
		{
			struct idx_struct newstruct;
			newstruct.used= FALSE;
			add_to_Chaine(data->indexes, &newstruct);
		}
	return num;
}

static int
SESH_open_db(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc,
						 CONST char ** argv)
{
	int numero;
	char num[10];
	struct local_data *data;
	int blocksize= 128;
	int i;
	int opening_mode = O_RDWR;

	data= (struct local_data *) clientData;


	if (argc < 2)
		return TCL_ERROR;
	numero= 0;
	while(numero < size_of_Chaine(data->files) && DBF(numero).used)
		numero++;
	if (numero >= size_of_Chaine(data->files))
	{
		struct dbf_struct newstruct;
		newstruct.used= FALSE;
		add_to_Chaine(data->files, &newstruct);
	}
	
	for (i= 2; i < argc; i++)
		{
			switch (*argv[i]) 
				{
				case 'r' :
					opening_mode= O_RDONLY;
					break;
				case 'w':
					opening_mode= O_RDWR;
					break;
				default:
					if (isdigit(*argv[i])) {
						int taille= atoi(argv[i]);
						if (taille > 10 && taille <= 256) {
							blocksize= taille;
						}
					}
				}
		}
	if (btdb_open(&(DBF(numero).file),
								argv[1], 
								opening_mode,
								blocksize,
								0666) != -1)
		{
			
			DBF(numero).used= TRUE;
			sprintf(num, "%d", numero);
			Tcl_AppendResult(interp, num, (char *) NULL);
			return TCL_OK;
		}
	else {
		Tcl_AppendResult(interp, "can't open database ", argv[1], (char *) NULL);
		return TCL_ERROR;
	}
}

static int
SESH_close_db(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc,
						 char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 2)
		return TCL_ERROR;
	numero= atoi(argv[1]);
	btdb_close(&(DBF(numero).file));
	DBF(numero).used= FALSE;
	return TCL_OK;
}

static void 
/*show_node(pstring *key, pstring *data, char* ptrs[2])*/
show_node(pstring *key, pstring *data, void *extdata)
{	
	char** ptrs= extdata;
	Tcl_Obj * s;
	
	Tcl_Interp *interp = (Tcl_Interp *)ptrs[0];		
	Tcl_Obj *	list= (Tcl_Obj *)ptrs[1];

	s= Tcl_NewStringObj(key->data, key->length);
	Tcl_ListObjAppendElement(interp, list, s);
	
	s= Tcl_NewStringObj(data->data, data->length);
	Tcl_ListObjAppendElement(interp, list, s);
}

static int
SESH_dump_db(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc,
						 char ** argv)
{
	int numero;
	char * ptrs[2];
	Tcl_Obj * result;
	struct local_data *data;

	data= (struct local_data *) clientData;

	if (argc != 2)
		return TCL_ERROR;

	result= Tcl_NewListObj(0, 0);
	ptrs[0]= (char *)interp;
	ptrs[1]= (char *)result;

	numero= atoi(argv[1]);
	
	btdb_apply(&(DBF(numero).file),
						 show_node,
						 ptrs,
						 0);

	Tcl_SetObjResult(interp, result);
	return TCL_OK;
}


static int
SESH_debug_db(ClientData clientData,
							Tcl_Interp *interp,
							int argc,
							char ** argv)
{
	int numero;
	char * ptrs[2];
	Tcl_Obj * result;
	struct local_data *data;

	data= (struct local_data *) clientData;

	if (argc != 2)
		return TCL_ERROR;

	result= Tcl_NewListObj(0, 0);
	ptrs[0]= (char *)interp;
	ptrs[1]= (char *)result;

	numero= atoi(argv[1]);
	
	btdb_dump(&(DBF(numero).file));
	
	Tcl_SetObjResult(interp, result);
	return TCL_OK;
}


static char *
pstring_to_char(pstring content)
{
	char *s;
	s= (char *)ALLOUER(sizeof(char)*(unsigned)(1+content.length));
	memcpy(s, content.data, (size_t)content.length);
	s[content.length]= 0;
	return s;
}

static int
SESH_get_db(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc,
						 char ** argv)
{
	int numero;
	pstring key, content;
	char *s;
	struct local_data *data;

	data= (struct local_data *) clientData;

	if (argc != 3)
		return TCL_ERROR;
	numero= atoi(argv[1]);
	if(NODBF(numero))
	{
		Tcl_AppendResult(interp, 
										 "Bad database ", argv[1],
										 (char *) NULL);
		return TCL_ERROR;
	}
	else
	{
		key.data= (char*)argv[2];
		key.length= strlen(argv[2]);
		
		if (btdb_find(&(DBF(numero).file), &key, &content))
			{
				s= pstring_to_char(content);
				LIBERE(content.data);
				Tcl_AppendResult(interp, s, (char *) NULL);
				LIBERE(s);
				return TCL_OK;
			}
		else
			Tcl_AppendResult(interp, "", (char *) NULL);
			return TCL_OK;
	}
}

static int
SESH_put_db(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc,
						 char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 4)
		return TCL_ERROR;
	
	numero= atoi(argv[1]);
	if(NODBF(numero))
	{
		Tcl_AppendResult(interp, 
										 "Bad database ", argv[1],
										 (char *) NULL);
		return TCL_ERROR;
	}
	else
	{
		btdb_insert_strings(&(DBF(numero).file),
												argv[2],
												argv[3]);
		return TCL_OK;
	}
}

static int
SESH_delete_db(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc,
						 char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 3)
		return TCL_ERROR;
	
	numero= atoi(argv[1]);
	if(NODBF(numero))
	{
		Tcl_AppendResult(interp, 
										 "Bad database ", argv[1],
										 (char *) NULL);
		return TCL_ERROR;
	}
	else
	{
		/*  btdb_delete_node(); */

		btdb_delete_string(&(DBF(numero).file),
										 argv[2]);
		return TCL_OK;
	}
}

static int 
SESH_reorg_db(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc,
						 char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 2)
		return TCL_ERROR;

	numero= atoi(argv[1]);
	if(NODBF(numero))
	{
		Tcl_AppendResult(interp, 
										 "Bad database ", argv[1],
										 (char *) NULL);
		return TCL_ERROR;
	}
	else
	{
		Tcl_AppendResult(interp, 
										 "function not available ", argv[1],
										 (char *) NULL);
		return TCL_OK;
	}
}

static int 
SESH_firstkey_db(ClientData clientData,
								 Tcl_Interp *interp,
								 int argc,
								 char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 2)
		return TCL_ERROR;

	numero= atoi(argv[1]);
	if(NODBF(numero))
		{
			Tcl_AppendResult(interp, 
											 "Bad database ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;
		}
	else
		{
			Tcl_AppendResult(interp, 
											 "function not available ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;
			
			/*
				datum key;
				char *s;
				key= gdbm_firstkey(DBF(numero).file);
				if (key.dptr)
				{
				s= (char *)ALLOUER(sizeof(char)*(unsigned)(1+key.dsize));
				memcpy(s, key.dptr, (size_t)key.dsize);
				s[key.dsize]= 0;
				Tcl_AppendResult(interp, s, NULL);
				LIBERE(s);
				LIBERE(key.dptr);
				}
				else
				{
				Tcl_AppendResult(interp, "", NULL);
				}
				return TCL_OK;
			*/
		}
}


static int 
SESH_nextkey_db(ClientData clientData,
								 Tcl_Interp *interp,
								 int argc,
								 char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 3)
		return TCL_ERROR;
	numero= atoi(argv[1]);
	if(NODBF(numero))
		{
			Tcl_AppendResult(interp, 
											 "Bad database ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;
		}
	else
		{
			/* datum key, content;
				 char *s;
				 key.dsize= strlen(argv[2]);
				 key.dptr=  argv[2];
				 content= gdbm_nextkey(DBF(numero).file, key);
				 if (content.dptr)
				 {
				 s= (char *)ALLOUER(sizeof(char)*(unsigned)(1+content.dsize));
				 memcpy(s, content.dptr, (size_t)content.dsize);
				 s[content.dsize]= 0;
				 Tcl_AppendResult(interp, s, NULL);
				 LIBERE(s);
				 LIBERE(content.dptr);
				 }
				 else
				 Tcl_AppendResult(interp, "", NULL);
				 return TCL_OK;
			*/
			Tcl_AppendResult(interp, 
											 "function not available ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;

		}
}

static int 
first_db_idx(ClientData clientData,
							Tcl_Interp *interp,
							int argc,
							char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 2)
		return TCL_ERROR;

	numero= atoi(argv[1]);
	if(NODBF(numero))
		{
			Tcl_AppendResult(interp, 
											 "Bad database ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;
		}
	else
		{
			char s[10];
			int i= get_new_index(clientData);
			sprintf(s, "%d", i);
			IDX(i).used= TRUE;
			IDX(i).idx= btdb_first_index((&(DBF(numero).file))); 
			return TCL_OK;
		}
}


static int 
get_db_idx(ClientData clientData,
							Tcl_Interp *interp,
							int argc,
							char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 3)
		{
			Tcl_AppendResult(interp, argv[0],
											 "Bad number of arguments.", 
											 (char *) NULL);

			return TCL_ERROR;
		}

	numero= atoi(argv[1]);
	if(NODBF(numero))
		{
			Tcl_AppendResult(interp, 
											 "Bad database ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;
		}
	else
		{
			int i;
			pstring key;

			key.data= (char*)argv[2];
			key.length= strlen(argv[2]);

			i= get_new_index(clientData);

			if (! btdb_get_index_to(&(DBF(numero).file),
															&(IDX(i).idx),
															&key))
				{
					Tcl_AppendResult(interp, 
													 "no element is greater than or equal to ", argv[2],
													 (char *) NULL);
					IDX(i).used= FALSE;
					return TCL_ERROR;
				}
			else 
				{
					char s[10];
					sprintf(s, "%d", i);
					IDX(i).used= TRUE;
					Tcl_AppendResult(interp, s, (char *) NULL);
				}
			return TCL_OK;
		}
}

static int
getLT(ClientData clientData,
			Tcl_Interp *interp,
			int objc,
			Tcl_Obj *CONST objv[])
{
	/*
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 3 && argc != 4)
		{
			Tcl_AppendResult(interp, argv[0],
											 "Bad number of arguments.", 
											 (char *) NULL);

			return TCL_ERROR;
		}

	numero= atoi(argv[1]);
	if(NODBF(numero))
		{
			Tcl_AppendResult(interp, 
											 "Bad database ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;
		}
	else
		{
			int i;
			pstring key;

			key.data= (char*)argv[2];
			key.length= strlen(argv[2]);

			i= get_new_index(clientData);

			if (! btdb_get_index_to(&(DBF(numero).file),
															&(IDX(i).idx),
															&key))
				{
					Tcl_AppendResult(interp, 
													 "no element is greater than or equal to ", argv[2],
													 (char *) NULL);
					IDX(i).used= FALSE;
					return TCL_ERROR;
				}
			else 
				{
					char s[10];
					sprintf(s, "%d", i);
					IDX(i).used= TRUE;
					Tcl_AppendResult(interp, s, (char *) NULL);
				}
			return TCL_OK;
		}
	*/
	return TCL_OK;
}

static int 
incr_db_idx(ClientData clientData,
						Tcl_Interp *interp,
						int argc,
						char ** argv)
{
	struct local_data *data;
	int idnum, increment, success;
	data= (struct local_data *) clientData;

	if (argc < 2)
		{
			Tcl_AppendResult(interp, 
											 "Bad number of arguments : ", argv[0],
											 " should be INDEX ?INCR ",
											 (char *) NULL);
			return TCL_ERROR;
		}

	idnum= atoi(argv[1]);
	
	if(NOIDX(idnum))
		{
			Tcl_AppendResult(interp, 
											 "Bad index ", argv[2],
											 (char *) NULL);
			return TCL_ERROR;
		}
	
	if (argc == 3)
		increment= atoi(argv[2]);
	else
		increment= 1;
	
	
	if (increment > 0)
		{
			while (increment-- 
						 && (success= btdb_next_index(IDX(idnum).idx)));
		}
	else
		{
			while (increment++ < 0
						 && (success= btdb_previous_index(IDX(idnum).idx)));
		}
	if (success)
		Tcl_AppendResult(interp, "1", (char *) NULL);
	else
		Tcl_AppendResult(interp, "0", (char *) NULL);
	return TCL_OK;
}

static int
getfrom_db_idx(ClientData clientData,
							Tcl_Interp *interp,
							int argc,
							char ** argv)
{
	struct local_data *data;
	pstring key, content;
	int idnum;
	int pos= 1;
	int onlykey= FALSE;

	data= (struct local_data *) clientData;

	if (argc < 2)
		{
			Tcl_AppendResult(interp, 
											 "Bad number of arguments ", argv[0],
											 "getfrom_db_idx [-key] ?index",
											 (char *) NULL);
			return TCL_ERROR;
		}

	if (strcmp(argv[pos],"-key") == 0)
		{
			onlykey= TRUE;
			pos++;
		}
	
	idnum= atoi(argv[pos]);
	if(NOIDX(idnum))
		{
			Tcl_AppendResult(interp, 
											 "Bad index ", argv[pos],
											 (char *) NULL);
			return TCL_ERROR;
		} else {
			if (onlykey) {
				Tcl_Obj * result;
				btdb_index_to_value(IDX(idnum).idx, &key, NULL);
				result= Tcl_NewStringObj(key.data, key.length);
				LIBERE(key.data);
				Tcl_SetObjResult(interp, result);
			}
			else {
				void *ext[2];
				Tcl_Obj * result;
				result= Tcl_NewListObj(0, 0);
				btdb_index_to_value(IDX(idnum).idx, &key, &content);
				ext[0]= interp;
				ext[1]= result;
				show_node(&key, &content, ext);
				LIBERE(key.data);
				LIBERE(content.data);
				Tcl_SetObjResult(interp, result);
			}
		}
	return TCL_OK;
}

static int 
islast_db_idx(ClientData clientData,
							Tcl_Interp *interp,
							int argc,
							char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 2)
		return TCL_ERROR;

	numero= atoi(argv[1]);
	if(NODBF(numero))
		{
			Tcl_AppendResult(interp, 
											 "Bad database ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;
		}
	else
		{
			Tcl_AppendResult(interp, 
											 "function not available ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;
		}
}

static int 
free_db_idx(ClientData clientData,
							Tcl_Interp *interp,
							int argc,
							char ** argv)
{
	int numero;
	struct local_data *data;
	data= (struct local_data *) clientData;

	if (argc != 2)
		return TCL_ERROR;

	numero= atoi(argv[1]);
	if(NOIDX(numero))
		{
			Tcl_AppendResult(interp, 
											 "Bad index ", argv[1],
											 (char *) NULL);
			return TCL_ERROR;
		}
	else
		{
			IDX(numero).used= FALSE;
			btdb_free_index(IDX(numero).idx);
			return TCL_OK;
		}
}



EXPORT(int,Dbinter_Init)(interp)
		 Tcl_Interp *interp;
{
	struct local_data *data;

	data= new_local_data();
	Tcl_SetAssocData(interp, "DBINTER", delete_local_data, (ClientData)data);
	
	Tcl_CreateCommand(interp, "open_db", SESH_open_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "close_db", SESH_close_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "get_db", SESH_get_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "put_db", SESH_put_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "delete_db", SESH_delete_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);	
	Tcl_CreateCommand(interp, "reorg_db", SESH_reorg_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "firstkey_db", SESH_firstkey_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "nextkey_db", SESH_nextkey_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "dump_db", SESH_dump_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);	
	Tcl_CreateCommand(interp, "debug_db", SESH_debug_db,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);
	/*
	 * New commands
	 *
	 * getGT <VAL> [<NB>]
	 * getGE <VAL> [<NB>]
	 * getLT <VAL> [<NB>]
	 * getLE <VAL> [<NB>]
	 *
	 * get NB entries with key strictly greater, greater or equal,
	 * strictly lesser, lesser or equal to <VAL>.
	 * <NB> default value is 1.
	 *
	 *
	 */
	//Tcl_CreateObjCommand(interp, "seshdb::getGT", getGT,
	//(ClientData) data, (Tcl_CmdDeleteProc *) NULL);	
	
	Tcl_CreateCommand(interp, "first_db_idx", first_db_idx,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);	
	Tcl_CreateCommand(interp, "get_db_idx", get_db_idx,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);	
	Tcl_CreateCommand(interp, "incr_db_idx", incr_db_idx,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);	
	Tcl_CreateCommand(interp, "getfrom_db_idx", getfrom_db_idx,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);	
	Tcl_CreateCommand(interp, "free_db_idx", free_db_idx,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);	
	Tcl_CreateCommand(interp, "islast_db_idx", islast_db_idx,
										(ClientData) data, (Tcl_CmdDeleteProc *) NULL);	

	return Tcl_PkgProvide(interp, "Dbinter", "1.1");
	return TCL_OK;
}

/*
 * Local Variables:
 * tab-width: 2
 * End:
 */

