/*
 * Simple Bitmap drawing lib
 */

/*  2001 Serge Rosmorduc
 * rosmord@iut.univ-paris8.fr
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: SimpleBitmap.c,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 */

/*
 * $Log: SimpleBitmap.c,v $
 * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * Imported using tkCVS
 *
 */

/* 
 * See SimpleBitmap.h for more information about coordinate system.
 */
#include <stdlib.h>
#include "general.h"
#include "XPoint.h"
#include "SimpleBitmap.h"
#include "chaine.h"

static void
computeIntersections(Chaine tab, double y, int nbr, XPoint* pts);

static void
sort_by_x(Chaine chp);

SimpleBitmap *
SimpleBitmap_init(int w, int h)
{
  SimpleBitmap *o;

  o= NOUV(SimpleBitmap);
  o->w= w;
  o->h= h;
  o->wchar= (w+7)/8;
  o->memsize= o->wchar*o->h;
  o->d= (unsigned char*)ALLOUER(o->memsize*sizeof(unsigned char));
  memset(o->d, 0, o->memsize*sizeof(unsigned char));
  return o;
}

void
SimpleBitmap_free(SimpleBitmap *o)
{
  LIBERE(o->d);
  LIBERE(o);
}

void
SimpleBitmap_clean(SimpleBitmap *o)
{
  memset(o->d, 0, o->memsize*sizeof(unsigned char));
}

void
SimpleBitmap_iplot(SimpleBitmap *o, int x, int y, int BLACK)
{
  if (x < 0 || y < 0)
    return;
  
  if (x < o->w && y < o->h)
    {
      if (BLACK)
	o->d[y*o->wchar+(x/8)] |= 1<< (7 - (x % 8));
      else
	o->d[y*o->wchar+(x/8)] &= ~(1<<(7- (x % 8)));
    }
}

int
SimpleBitmap_getPixel(SimpleBitmap *o, int x, int y)
{
  if (x < 0 || y < 0 || x > o->w || y > o->h)
    return 0;
	else
		return o->d[y*o->wchar+(x/8)] & (1<<(7- (x % 8)));
}

/* 
 * Bresenham Algorithm, taken from 
 * GRAPHICS PROGRAMMING column, Lines Galore, by Kent Porter
 * in DDOBS Journal, 1989
 */

void
SimpleBitmap_iline(SimpleBitmap *o,
		   int x1,  int y1,
		   int x2,  int y2,
		   int black)
{
  int  w, h, d, dxd, dyd, dxn, dyn, dinc, ndinc, p;
  int x, y;
  
  /* Set up */
  x = x1; y = y1;                          /* start of line */
  w = x2 - x1;                      /* width domain of line */
  h = y2 - y1;                     /* height domain of line */
  
  /* Determine drawing direction */
  if (w < 0) {                     /* drawing right to left */
    w = -w;                               /* absolute width */
    dxd = -1;                    /* x increment is negative */
  } else                           /* drawing left to right */
    dxd = 1;                       /* so x incr is positive */
  if (h < 0) {                     /* drawing bottom to top */
    h = -h;                       /* so get absolute height */
    dyd = -1;                         /* y incr is negative */
  } else                           /* drawing top to bottom */
    dyd = 1;                       /* so y incr is positive */
  
  /* Determine major axis of motion */
  if (w < h) {                           /* major axis is Y */
    p = h, h = w, w = p;       /* exchange height and width */
    dxn = 0;
    dyn = dyd;
  } else {                               /* major axis is X */
    dxn = dxd;
    dyn = 0;
  }
  
  /* Set control variables */
  ndinc = h * 2;                  /* Non-diagonal increment */
  d = ndinc - w;                /* pixel selection variable */
  dinc = d - w;                       /* Diagonal increment */
  
  /* Loop to draw the line */
  for (p = 0; p <= w; p++) {
    SimpleBitmap_iplot(o, x, y, black);
    if (d < 0) {                     /* step non-diagonally */
      x += dxn;
      y += dyn;
      d += ndinc;
    } else {                             /* step diagonally */
      x += dxd;
      y += dyd;
      d += dinc;
    }
  }
}


/**
 * If we use Bresenham's algorithm to draw lines, we want the fill Polygon algorithm to be 
 * compatible with it. This isn't done yet.
 */

void
SimpleBitmap_fillPolygon(
			 SimpleBitmap *o,
			 int npoints, 
			 XPoint * lespoints,
			 int sens)
{
  int i, a;
  static Chaine chp= NULL;
  XPoint min, max;
  
  if (npoints==0) return;

  min=max= lespoints[0];
  
  /* First, compute bounding box */
  for (i=1; i< npoints; i++)
    {
      if (lespoints[i].x > max.x)
				max.x= lespoints[i].x;
      if (lespoints[i].x < min.x)
				min.x= lespoints[i].x;
      if (lespoints[i].y > max.y)
				max.y= lespoints[i].y;
      if (lespoints[i].y < min.y)
				min.y= lespoints[i].y;
    }
  
  for(a= min.y; a<= max.y; a++)
    {
      XPoint curr;
      curr.x=0;
      curr.y= a;
      
      
      if(! chp)
				init_Chaine(&chp, sizeof(XPoint));
      
      /* For each line
       * a) Compute intersections
       * b) sort them
       * c) Draw the corresponding lines
       */
       
      /* computation */
      computeIntersections(chp, curr.y, npoints, lespoints);
      /* sorting */
      sort_by_x(chp);
      /* drawing */
      for(i=0; i + 1< size_of_Chaine(chp); i+= 2)
				{
					XPoint p1, p2;
					p1= *((XPoint*)get_from_Chaine(chp, i));
					p2= *((XPoint*)get_from_Chaine(chp, i+1));
					if  (sens == 0)
						{
							/* inked parts should not be erased by tangent 
								 uninked parts.
								 
								 REM: we suppose that p1.x, p1.y is black.  If it
								 isn't, painting it in white would have been a NO OP
								 operation anyway.
							*/

							
							while(p1.x < p2.x && SimpleBitmap_getPixel(o, p1.x - 1, p1.y) == 0)
								{ 
									p1.x++;
								}
							
							while(p1.x < p2.x && SimpleBitmap_getPixel(o, p2.x + 1, p2.y) == 0)
								{
									p2.x--;
								}
						}
					if (p1.x < p2.x)
						SimpleBitmap_iline(o, p1.x, p1.y, p2.x, p2.y, sens);
				}
      
      reset_Chaine(chp);
    }
  /* Contour ? */
  /*    for(i=1; i < npoints; i++) */
  /*      { */
  /*        SimpleBitmap_iline(o, lespoints[i-1].x, lespoints[i-1].y, 
						lespoints[i].x, lespoints[i].y, sens); */
  /*      } */
  /*    if (npoints > 0) { */
  /*      SimpleBitmap_iline(o, lespoints[npoints-1].x, lespoints[npoints-1].y,  */
  /*  		       lespoints[0].x, lespoints[0].y, sens); */
  /*    } */
}

/**
 * Utility functions for polygon filling
 */

static void
computeIntersections(Chaine tab, double y, int nbr, XPoint* pts)
{
  int i;
  XPoint old= pts[nbr-1];

  y+= 0.000001;
  for (i=0; i< nbr; i++)
    {
      if (((pts[i].y >= y) && (old.y < y))
					|| ((pts[i].y <= y) && (old.y > y)))
				{
					/* Calculer et rajouter l'intersection */
					XPoint p;
					double t= (y-old.y)/(pts[i].y-old.y);
					p.y= y;
					/* At least round x to nearest value ! */
					p.x= (t*pts[i].x + (1-t)*old.x);
					add_to_Chaine(tab,&p);
				}
      old= pts[i];
    }
}

#define GFC(i) (*(XPoint *)get_from_Chaine(chp,i))

static void
sort_by_x(Chaine chp)
{
  int trie= FALSE;
  int taille= size_of_Chaine(chp);

  while (!trie)
    {
      int i;
      trie= TRUE;
      for (i= 0; i< taille - 1; i++)
	if (GFC(i).x > GFC(i+1).x)
	  {
	    XPoint p;
	    trie= FALSE;
	    p= GFC(i);
	    GFC(i)= GFC(i+1);
	    GFC(i+1)= p;
	  }
    }
}

/*
 * Dump functions :
 */

/**
 * SimpleBitmap_toPBMFile --
 *	 dumps to a file
 */

void
SimpleBitmap_toPBMFile(FILE *f, SimpleBitmap *o)
{
  int i;
  fprintf(f, "P4\n");
  fprintf(f, "%d %d\n", o->w, o->h);
  for (i= 0; i< o->memsize ; i++)
    {
      fprintf(f, "%c", (int) o->d[i]);
    }
  fprintf(f, "\n");
}



/*
 * Local Variables:
 * tab-width: 2
 * End:
 */
