# dicoTk.tcl --
#
#       
#	Dictionnary definition and interface for
#       the tksesh program. Tag oriented stuff with indexes
#       all over the place
#
# 
#
# Representation :
# A dictionary is a structure which contains :
#
# a main file and a list of indexes
# a max record number
#
# each record is indexed by a number (the max current number)
# there are indexes on a number of fields
# each index provides a LIST of the record numbers that contains the given
# value for the index
#
# each record in the main file has the following structure : 
# it is TAG oriented and built with lists.
# each TAG has the following form :
# {TAGNAME VALUE} 
# VALUE is mandatory. Its structure depends on the TAG. 
# an empty value is represented by {}.
#
#
#
# *************************************************************************
#
# In dumping, a tagon is replaced by "{TAGNAME" ... 
# and a tagoff is replaced by "}"
#
#  
#
#
# So, we should have now the following tags ::
# GRP : basic group container
# WRI : possible written form of the word
# TRAD : translation
# MTRI : translitteration du mot
# TRI  : texte en translitteration
# COM : a comment. The first area of the comment is the type of the comment, the next one 
#       is free text, containing both WRI, TRL, and TRI, texts. 
#       The text in COM is not currently indexed. It could be, but in his own index
# REF : a reference to a text
# AREF : reference to an article, or, generally, a written presentation
# GRA : Grammar
# HIE : hieroglyphic text.
# COP : coptic text.
# USE : usage
# PCDATA : Untagged text. 
#
#
# Each TAG is described by the following data :
#
# indentation
# above-line indentation
# begin-string 
# end-string
# 
#
# Now, Some tags are legal in certain contexts and some other are not.
# Legal tags are :
# 
# CTXT       | LEGAL TAGS
# ___________|______________________________________________
# GRP        | WRI TRAD TRI COM REF AREF GRP USE GRA(Not PCDATA !)
# WRI        | Nothing. Special editor.
# HIE        | Nothing. Special editor.
# COM        | All
# TRAD       | PCDATA
# TRI        | PCDATA
# REF        | PCDATA
# AREF       | PCDATA
# USE        | HIE, TRI
# GRA        | Nothing for the moment
#
# dico.dtd is the reference for the final representation
#
# Structure for the database
#
# The records are in DICONAME.
# A special record is MAX, which contain the maximum id
#
# All indexes are in bases called DICONAME.TAG,
# where TAG is the indexed TAG.
#

package require legendes
# Moving toward sqlite db.
# package require seshbase

package require hieroglyphs
package require seshutils
package require seshw
package require seshnormalize
package require Dbinter


package provide seshDico 0.0	;# This uses Tcl namespaces

package require myStoopWidgets


class DicoTk {

    proc DicoTk {this parent name type} SFrame {$parent $name} {
	set ($this,wf) [new DicoWF $type]
    }

    proc ~DicoTk {this} {
	PRIVUnregister $this 
    }

    proc getTextPath {this} {
	set p [Component::getPath $this]
	return $p.t
    }
    
    # remember this object's address, knowing its path.
    proc PRIVRegister {this} {
	set p [getTextPath $this]
	set (path2Object,$p) $this
    }
    
    proc PRIVUnregister {this} {
	set p [getTextPath $this]
	if [info exists (path2Object,$p)] {
	    unset (path2Object,$p) 
	}
    }
    
    # Help procedure, used for widget bindings.
    # Basically, we know the widget's path and we want to retrieve the object.
    proc callWindowCommand {command path args} {
	set obj $(path2Object,$path)
	eval $command $obj $args
    }
    
    virtual proc creatorHook {this} {
	# creates a dictionary window
	set path [Component::getPath $this]

	# menus
	frame $path
	frame $path.menus
	
	menubutton $path.menus.file -text [legendes::trans "Card"] -menu $path.menus.file.m	
	menubutton $path.menus.edit -text [legendes::trans "Edit"] -menu $path.menus.edit.m

	menu $path.menus.file.m
	menu $path.menus.edit.m
	label $path.menus.tag -textvariable DicoTk::VARS($path,upperTag)

	pack $path.menus.file $path.menus.edit -side left
	pack $path.menus.tag -side right
	
	$path.menus.file.m add command -label [legendes::trans "New"] -command "DicoTk::clearWindow $this"
	$path.menus.file.m add command -label [legendes::trans "Save"] -command "DicoTk::save $this"
	$path.menus.file.m add command -label [legendes::trans "Save as HTML file"] -command "DicoTk::save_as_html $this"
	$path.menus.file.m add command -label [legendes::trans "Save as LaTeX file"] -command "DicoTk::save_as_latex $this"
	#$path.menus.file.m add command -label [legendes::trans "Close"] -command "DicoTk::close $this"
	
	
	$path.menus.edit.m add command -label [legendes::trans "Select current group"] \
	    -command [list DicoTk::selectTag $this]

	$path.menus.edit.m add command -label [legendes::trans "Delete selected group"] \
	    -command [list DicoTk::deleteSelected $this]

	$path.menus.edit.m add command -label [legendes::trans "copy selected group"] \
	    -command [list DicoTk::copy $this]
	
	$path.menus.edit.m add command -label [legendes::trans "Paste"]\
	    -command [list DicoTk::paste $this]

	$path.menus.edit.m add command -label [legendes::trans "Paste Text"]\
	    -command [list DicoTk::pastetext $this]

	frame $path.entete 

	#text window and scollbars
	text $path.t \
	    -yscrollcommand "$path.y set" \
	    -xscrollcommand "$path.x set"\
	    -wrap word\
	    -font {times 14}

	scrollbar $path.x -command "$path.t xview" -orient horizontal
	scrollbar $path.y -command "$path.t yview" -orient vertical
	bindtags $path.t [list $path.t HierarchicText all] 

	# grid
	grid $path.menus -sticky ew
	grid $path.t $path.y -sticky news
	grid $path.x -sticky ew
	grid rowconfigure $path 1 -weight 1
	grid columnconfigure $path 0 -weight 1
	
	# useful stuff for the text window :(TO SUPPRESS ?)
	$path.t tag configure begin -foreground red -spacing1 10
	$path.t tag configure end -foreground red -spacing3 10
	$path.t tag configure translit -foreground blue -font {times 14 italic}

	$path.t tag configure blue -background lightblue
	
	clearWindow $this
	PRIVRegister $this
    }
    
    proc clearWindow {this} {
	# Clear the workflow.
	DicoWF::resetWF $($this,wf)
	# Actual deletion of the window content
	set path [Component::getPath $this]
	$path.t delete 0.0 end
	# Reset the 
	# number of sub windows in this window
	set ($this,subWinNbr) 0
    }


    # Load entry "id" from the base.
    proc load {this id} {
	set wf $($this,wf)
	DicoWF::setEntry $wf $id
	set txt [DicoWF::getEntryText $wf]
	setText $this $txt
    }

    proc close {this} {
    }
    
    # DicoTk::save --
    #          save the content of window w to the disk
    # Arguments:
    #   w : window to save
    #
    # Results:
    #   no result returned.
    #   the data is saved
    #   the indexes are set to their proper values :
    #     * new data is entered
    #     * old data is cleared
    
    proc save {this} {
	set txt [asText $this 0.0 end 1]
	DicoWF::saveEntry $($this,wf) $txt
    }

    ## Fill a dictionary entry widget according to the text :
    # setText returns 1 if the text inserted requires a white space after it,
    # and else 0

    proc setText {this txt {all 1}} {
	if {$all} {
	    clearWindow $this
	}
	
	set space ""
	set wantspace 0
	
	set path [Component::getPath $this]
	if {[TagHandler::isTag $txt]} {
	    set tagName [lindex $txt 0]
	    set tagData [lindex $txt 1]
	    
	    if {[TagHandler::hasLoadFunction $tagName]} {
		[TagHandler::getLoadFunction $tagName] $this [deprotect $tagData]
		set wantspace 0
	    } else {
		if {! $all } {
		    set currenttag [insertTag $this $tagName]
		}
		foreach i $tagData {
		    $path.t insert insert $space
		    if [setText $this $i 0] {
			set space " "
			set wantspace 1
		    } else {
			set space ""
			set wantspace 0
		    }
		}
		if {! $all } {
		    set end [lindex [$path.t tag prevrange $currenttag insert] 1]
		    $path.t mark set insert $end
		}
		set wantspace 0
	    }
	} else {
	    $path.t insert insert [deprotect $txt]
	    set wantspace 1
	}
	return $wantspace
    }

    ################################################################################
    #
    # Editing code.
    #
    
    proc deleteSelected {this} {
	set path [Component::getPath $this]
	$path.t delete sel.first sel.last
    }

    proc isTextProtected {this pos} {
	set path [Component::getPath $this]
	return [expr [lsearch -regexp [$path.t tag names $pos] {begin_.*|end_.*}] != -1]
    }

    proc textdelete {this} {
	set path [Component::getPath $this]
	if {! [isTextProtected $this insert] && [$path.t compare "insert + 1 chars" != end]} {
	    $path.t delete insert
	}
    }

    proc textbackspace {this} {
	set path [Component::getPath $this]
	if {! [isTextProtected $this "insert - 1 chars"]} {
	    $path.t delete "insert - 1 chars"
	}
    }
    
    # Is it possible to insert PCDATA at the current position ?
    # It is not possible to insert a character IN a "begin" or "end" tag.
    # This means it is not possible to insert a character if both the current AND
    # the previous characters are in the same begin or end.
    # That is, TAGS(insert) INTER TAGS(insert - 1 chars) does not contain 
    # a begin or end tag

    proc canInsertText {this} {
	return [expr ! [isNotLegalAtInsertPoint $this PCDATA]]
    }
    
    # Compute the char to add, knowing the value and
    # the current tag. This allows different mapping for different languages
    # e.g. Coptic.

    proc keyToValue {this key} {
	set tag [TagHandler::getTagName [getUpperTag $this]]
	if {[TagHandler::hasKeyBinding $tag $key]} {
	    return [TagHandler::getKeyBinding $tag $key]
	} else {
	    return $key
	}
    }

    proc keyinsert {this key} {
	textinsert $this [keyToValue $this $key]
    }


    proc textinsert {this string} {
	if [canInsertText $this] {
	    set path [Component::getPath $this]
	    $path.t insert insert $string
	    PRIVupdatePosition $this
	}
    }

    proc copy {this} {
	clipboard clear
	set txt [asText $this sel.first sel.last]
	set txt [lindex [lindex $txt 1] 0]; # cadr
	clipboard append  $txt
    }

    proc paste {this} {
	set txt [selection get -selection CLIPBOARD]
	set tag [lindex $txt 0]
	if [isNotLegalAtInsertPoint $this $tag] {
		tk_dialog  .dialog  "error" "You can't insert this text here" "" 0 "ok"	    
	} else {
	    setText $this $txt 0
	}
    }

    proc winpastetext {this} {
	pastetext $this
	set path [Component::getPath $this]
	PRIVupdatePosition $this
    }

    proc pastetext {this} {
	global env
	set txt [selection get -selection $env(DEFAULTSELECTION)]
	if [isNotLegalAtInsertPoint $this PCDATA] {
		tk_dialog  .dialog  "error" "You can't insert this text here" "" 0 "ok"	    
	} else {
	    textinsert $this $txt
	}
    }
    
    proc buildTagSelectionMenu {this x y} {
	# Build a menu.
	set m .dicopop
	catch {destroy $m}
	set wf $($this,wf)
 	menu $m
	# get the context tag :
	set contextTag [TagHandler::getTagName [getUpperTag $this]]
	if { $contextTag == "" } {
	    set contextTag [DicoWF::getRootTag $wf]
	}
	
	# fill the menu accordingly

 	foreach subTag [TagHandler::getLegalTags $contextTag] {
	    if { $subTag != "PCDATA" } {
		if {[TagHandler::hasLoadFunction $subTag]} {
		    $m add command -label [TagHandler::getTagDescription $subTag]\
			-command [list [TagHandler::getLoadFunction $subTag]  $this ]
		} else {
		    $m add command -label [TagHandler::getTagDescription $subTag]\
			-command [list DicoTk::insert_parenthesis $this $subTag ]
		}
	    }
	}

	# Post the menu
	catch {::tk_popup $m $x $y }
    }
    
    proc textleft {this} {
	PRIVmoveInsertMark $this "insert  - 1 chars" 
    }

    proc textright {this} {
	PRIVmoveInsertMark $this "insert  + 1 chars" 
    }

    proc textup {this} {
	PRIVmoveInsertMark $this "insert  - 1 lines"
    }

    proc textdown {this} {
	PRIVmoveInsertMark $this "insert  + 1 lines"
    }


    proc seeRef {this} {
	set t [getUpperTag $this]

	if [string match REF* $t] {
	    set lims [[getTextPath $this] tag prevrange $t insert]
	    set txt [[getTextPath $this] get [lindex $lims 0] [lindex $lims 1]]
	    if [regexp {\[(.*)@(.*)\]} $txt dum text coords] {
		::Tksesh::showText $text $coords
	    }
	}
    }

    proc setBlue {this} {
	catch {
	    set path [Component::getPath $this]
	    set ntag [getUpperTag $this current]
	    set lims [$path.t tag prevrange $ntag current]
	    $path.t tag add blue [lindex $lims 0] [lindex $lims 1]
	}
    }

    proc unsetBlue {this} {
	set path [Component::getPath $this]
	catch {$path.t tag remove blue 0.0 end}
    }

    proc PRIVupdatePosition {this} {
	set path [Component::getPath $this]
	$path.t see insert
	# Refreshes the upper tag data.
	getUpperTag $this
    }

    proc PRIVmoveInsertMark {this where} {
	set path [Component::getPath $this]
	$path.t mark set insert $where
	PRIVupdatePosition $this
    }
    
    ############################################################
    # General utilities functions. Consider lists as sets.

    proc normalizeHieroglyphs {hi} {
	puts "working on \"$hi\""
	catch {sesh .tmp_s} err
	.tmp_s delete 0.0 end
	.tmp_s insert insert $hi
	set end [.tmp_s index insert]
	.tmp_s mark set insert 0.1
	set curr [.tmp_s index insert]
	set signs [.tmp_s liste_signes 0.0 end]
	set result {}
	regsub -all -- {&} $signs " " signs
	# Not split. Split is fussy about successive spaces
	return [hieroglyphs::normalizeText $signs 1]
    }
    
    ################################################################################
    #
    # CODE FOR DUMPING THE WIDGET IN VARIOUS FORMATS.
    #
    # Note that this code doesn't really belong in DicoTk.
    # However, currently, it uses resources from the DicoTk widget, in particular
    # when bitmaps for the hieroglyphs are required.
    
    proc getHieroAsText {seshw} {
	return [$seshw get 0.0 end]
    }
    
    proc getHieroAsHTML {seshw dir num} {
	set fname img$num.xbm
	set f [open [file join $dir $fname] w]
	puts $f [$seshw undocumented4 -xbm]
	close $f
	return [format {<IMG src="%s">} $fname]
    }

    proc getHieroAsLatex {seshw} {
	return [format {
	    \begin{hieroglyph}
	    %s
	    \end{hieroglyph}
	} [$seshw get 0.0 end]]
    }

    proc protect {txt} {
	set txt [string map {
	    %  %%
	    \{  %par
	    \} %rap
	    \" %quo
	} $txt]
	return $txt
    }

    proc deprotect {txt} {
	set txt [string map {
	    
	} $txt]
	regsub -all -- {%%} $txt {} txt
	regsub -all %par $txt \{ txt
	regsub -all %rap $txt \} txt
	regsub -all %quo $txt \" txt
	#regsub -all %_ $txt { } txt
	regsub -all -- {}  $txt {%} txt
	return $txt
    }

    
    # The dump command of the text widget sends back 
    # a number of irrelevant text separations. We 
    # deal with these.
    #
    # The selected text shall always cover a complete tag contents
    # PRIVdumpText will only send back :
    #
    # - tag limits (using "begin" tags)
    # - actual text (suppressing the text contains in tag limits)
    # - windows information.
    #
    # Note that a condition for this to work is that any tag limit is at least one char 
    # long. This is possible. We can protect the tag limits and even make some of them invisible.

    
    proc PRIVdumpText {this {begin 0.0} {end end}} {
	set path [Component::getPath $this]

	set txt [$path.t dump -text -tag -window $begin $end]
	set last {}

	# First, concatenate all text

	foreach {k v i} $txt {
	    switch --  $k {
		text {
		    append last $v
		}
		default {
		    if {[string compare $last ""] != 0} {
			lappend res text $last 0
			set last {}
		    }
		    lappend res $k $v $i
		}
	    }
	}

	# See what to do with the last element.
	if {[string compare $last ""] != 0} {
	    lappend res text
	    lappend res $last
	    lappend res 0
	    set last {}
	}
	return $res
    }

    
    proc asText {this {begin 0.0} {end end} {all 0}} {
	# save is used to mark text when we don't want to save it
 	set save 1
	set result "" 
	set space ""

	set txt [PRIVdumpText $this $begin $end]
	
	foreach {key value index} $txt {
	    switch -exact -- $key {
		text {
		    if {$save} {
			append result $space
			set space " "
			append result [protect $value]
		    }
		}
		tagon {
		    switch -glob -- $value {
			begin_* {
			    regexp {^begin_([^_]*)_} $value tmp tagname
			    append result "$space\{$tagname \{"
			    set space ""
			    set save 0
			}
			end* {
			    set save 0
			}
			default {}
		    }
		}
		tagoff {
		    switch -glob -- $value {
			begin_* {
			    set save 1
			}
			end* {
			    set save 1
			    append result "\}\}"
			    set space " "
			}
			default {}
		    }
		}
		window {
		    regexp {\.s([^0-9]+)[0-9]+} $value dum tagname
		    append result $space
		    set space " "
		    append result [list [TagHandler::buildTaggedText \
					     $tagname \
					     [protect [[TagHandler::getTagSaveProc $tagname] \
							   $value]]]]
		} 
		default {}
	    }
	}
	
	# Due to the tagging system and to the dump command, we can :
	#    * have extra tagoff command at the beginning of the text
	#    * miss tagoffs at the end
	# So, we should :
	# \{(1) suppress any trailing '\}'
	# \{(2) balance the string with the necessary '\}'
	# thanks to protect, it should work.
	
	if {! $all } {
	    regsub "^\}* *" $result "" result
	    regsub -all "\[^\}\]" $result "" a
	    regsub -all "\[^\{\]" $result "" b
	    set n [expr [string length $b] - [string length $a]]
	    for {set i 0} {$i < $n} {incr i} {
		append result "\}"
	    }
	}
	
	set result [TagHandler::buildTaggedText [DicoWF::getRootTag $($this,wf)] $result]
	return $result
    }

    # the following function will dump a page 
    # as a html document. 
    # When tags are provided for a window, they will be generated.
    # hieroglyphs will be dumped as graphic files.

    proc save_as_html {w} {
	set fic [tk_getSaveFile]
	if {[string compare $fic ""] != 0} {
	    dump_as_html $path $fic
	}
    }

    proc save_as_latex {w} {
	set fic [tk_getSaveFile]
	if {[string compare $fic ""] != 0} {
	    dump_as_latex $path $fic
	}
    }

    proc dump_as_html {w dirname} {
	variable THE_TAGS
	variable VARS
	set imgnum 0
 	#Save the window -- a text window
 	#First, get the text for the whole window.
	
 	# save is used to mark text when we don't want to save it
	
	file mkdir $dirname

 	set save 1
	set result "<HTML>\n<BODY>\n"
	set space ""

	set txt [PRIVdumpText $this 1.0 end]
	foreach {key value index} $txt {
	    switch -exact -- $key {
		text {
		    if {$save} {
			append result $space
			set space " "
			append result $value
		    }
		}
		tagon {
		    switch -glob -- $value {
			begin_* {
			    regexp {^begin_([^_]*)_} $value tmp tagname
			    if [info exists THE_TAGS($tagname,htmlbegin)] {
				append result $THE_TAGS($tagname,htmlbegin)
			    }
			    set space ""
			    set save 0
			}
			end* {
			    set save 0
			}
			default {}
		    }
		}
		tagoff {
		    switch -glob -- $value {
			begin_* {
			    set save 1
			}
			end* {
			    set save 1
			    regexp {^end_([^_]*)_} $value tmp tagname
			    if [info exists THE_TAGS($tagname,htmlend)] {
				append result $THE_TAGS($tagname,htmlend)
			    }
			    set space " "
			}
			default {}
		    }
		}
		window {
		    regexp {\.s([^0-9]+)[0-9]+} $value dum tagname
		    set fonc $THE_TAGS($tagname,htmlsave)
		    if {[string compare $fonc ""] != 0} {
			incr imgnum
			append result [$fonc $path $dirname $imgnum $value]
			set space " "
		    }
		} 
		default {}
	    }
	}
	append result "</BODY>\n</HTML>\n"
	set f [open [file join $dirname index.html] w]
	puts $f $result
	close $f
    }

    proc dump_as_latex {w file} {
	variable THE_TAGS
	variable VARS
	set imgnum 0
 	#Save the window -- a text window
 	#First, get the text for the whole window.
	
 	# save is used to mark text when we don't want to save it

 	set save 1
	set result {
\documentclass{article}
\usepackage{egypto}
\usepackage{hiero}
\usepackage[latin1]{inputenc}
\usepackage[T1]{fontenc}

\begin{document}
	}
	
	set space ""

	set txt [PRIVdumpText $this 1.0 end]
	foreach {key value index} $txt {
	    switch -exact -- $key {
		text {
		    if {$save} {
			append result $space
			set space " "
			append result $value
		    }
		}
		tagon {
		    switch -glob -- $value {
			begin_* {
			    regexp {^begin_([^_]*)_} $value tmp tagname
			    if [info exists THE_TAGS($tagname,latexbegin)] {
				append result $THE_TAGS($tagname,latexbegin)
			    }
			    set space ""
			    set save 0
			}
			end* {
			    set save 0
			}
			default {}
		    }
		}
		tagoff {
		    switch -glob -- $value {
			begin_* {
			    set save 1
			}
			end* {
			    set save 1
			    regexp {^end_([^_]*)_} $value tmp tagname
			    if [info exists THE_TAGS($tagname,latexend)] {
				append result $THE_TAGS($tagname,latexend)
			    }
			    set space " "
			}
			default {}
		    }
		}
		window {
		    regexp {\.s([^0-9]+)[0-9]+} $value dum tagname
		    set fonc $THE_TAGS($tagname,latexsave)
		    if {[string compare $fonc ""] != 0} {
			incr imgnum
			append result [$fonc $path $imgnum $value]
			set space " "
		    }
		} 
		default {}
	    }
	}
	append result {\end{document}}
	set f [open $file w]
	puts $f $result
	close $f
    }



    ################################################################################
    #
    # TAG ORIENTED STUFF.
    #
    #
    
    # getUpperTag returns the innermost tag of a specific kind at some place
    proc getUpperTag {this {pos insert}} {
    	set ltags [getTagsAtPos $this $pos]
    	regsub -all -- "(begin|end)_(\[^ \]+)" $ltags {} ltags
	regsub -all -- "((\[^ \]+)_(\[^ \]+))|(\[^ \]+)" $ltags {\1} ltags

	set ltags [lsort -command TagHandler::compareTags $ltags]
	set ($this,upperTag) [lindex $ltags end]
	return $($this,upperTag)
    }

    # return a number which is the embedding level of the insert point.

    proc getEmbeddingLevelAtPos {this tag pos} {
	set ltags [getTagsAtPos $this]
	# build a string containing only the tags associated numbers, which are
	# in fact tag depth.
	regsub -all -- ".*_(\[0-9\]+)|(\[^ \]+)|( )" $ltags {\1\3} ltags
	# The string contains only integers and spaces, and thus is a legal
	# tcl list.
	#
	# add depth 0 to the list, to ensure it's not empty.
	lappend ltags 0
	set ltags [lsort -integer  -decreasing $ltags]
	# get the maximal depth at point for a new tag.
	return [expr [lindex $ltags 0] + 1]
    }


    # isEmbedded returns 1 if the insertion point is inside a 
    # tagged area declared as "atomic", which means an internal window.
    
    proc isEmbedded {this} {
	set ltags [getTagsAtPos $this]
	return [TagHandler::isInAtomicTag $ltags]
    }

    proc isNotLegalAtInsertPoint {this tag} {
	set wf $($this,wf)

	set roottag [DicoWF::getRootTag $wf]

	set tags [concat ${roottag}_0 [getTagsAtPos $this]]
	
	return [expr {! [TagHandler::isLegalTagInTags $tag $tags]}]
    }

    # DicoTk::insertTag --
    #          Insert a tag at insertion point
    # Arguments:
    #	w	        name of the dico window
    #	tag	        tag to add
    #
    # Results:
    #        Returns the name of the created tag (the base tag name + some number)
    #        the markers for tag begin and tag end are inserted
    #        the insertion point is set at the middle of the tag
    
    proc insertTag {this tag} {
	set path [getTextPath $this]

	set l [getEmbeddingLevelAtPos $this $tag insert]
	set level $l
	set tagBody [TagHandler::buildTagBody ${tag} $l]
	set beginname [TagHandler::buildTagBegin ${tag} $l]
	set endname [TagHandler::buildTagEnd ${tag} $l]

	eval $path tag configure $tagBody [TagHandler::getTagConfiguration $tag body]
	eval $path tag configure $beginname [TagHandler::getTagConfiguration $tag begin]
	eval $path tag configure $endname [TagHandler::getTagConfiguration $tag end]

	foreach i [TagHandler::getTagBindings $tag] {
	    $path tag bind $tagBody [lindex $i 0] [lindex $i 1]
	}
	
	set context [getTagsAtPos $this]

	set begin [$path index insert];	# beginning of the tagged region	
	$path insert $begin [TagHandler::getBeginText $tag] $context;	# beginning text
	set begintext [$path index insert]; # text begin

	$path insert insert [TagHandler::getMiddleText $tag] $context
	set endtext [$path index insert]	
	$path insert insert [TagHandler::getEndText $tag] $context; # insert at end of text
	$path tag add $beginname $begin $begintext; # tag the beginning part
	$path tag add $endname $endtext insert
	$path tag add $tagBody $begin insert
	set end [$path index insert]	
	if {[TagHandler::isTagInsertAtEnd $tag]} {
	    $path mark set insert "$end"
	} else {
	    $path mark set insert "$begintext"
	}
	return $tagBody
    }

    # insert_parenthesis is here to allow the insertion of
    # parenthesized constructs.
    #   w : the text window
    #   tag : base name of the tag
    # Options :
    #  -begin, -middle, -end : the text to be inserted at the begin, middle, and end of the tag
    #  -conf  : configuration options for the tag.
    #          you can use the variable "level", which is the degree of embedding of the tag
    #  -bconf, -econf : special options for tag beginning and ending. Can be used to introduce vertical spaces
    #
    # returns the created tag, and {} if it wasn't created

    proc insert_parenthesis {this tag} {
	if {[isNotLegalAtInsertPoint $this $tag]} {
	    tk_dialog  .dialog  [legendes::trans "Beware"] \
		[legendes::trans "You can't insert %s here" $tag] \
		"" 0 [legendes::trans "Ok"]
	    return {}
	}
	insertTag $this $tag
    }

    proc insertHieroglyphicTag {this tag {txt ""}} {
	incr ($this,subWinNbr)
	set path [Component::getPath $this]
	set s $path.s$tag$($this,subWinNbr)
	::seshw::create_passive $s
	$path.t window create insert -window $s
	if { $txt ne "" } {
	    ::seshw::set_text $s $txt
	    ::seshw::autosize $s
	}
	bind $s <Double-Button-1> [list DicoTk::edit_hieroglyphs $this $s]
    }
 
    proc edit_hieroglyphs {this subwin} {
	set ed .dicohieredit
	toplevel $ed
	wm title $ed [legendes::trans "hieroglyphic word editor"]
	seshw::create_editor $ed.s -notitle
	grid $ed.s -
	$ed.s.s insert 0.0 [$subwin get 0.0 end]
	button $ed.ok -text [legendes::trans "Ok"]\
		-command [list DicoTk::setHieroFromEdit $subwin $ed]
	button $ed.cancel -text [legendes::trans "Cancel"]\
		-command [list destroy $ed]
	grid $ed.ok $ed.cancel
    }
    
    proc setHieroFromEdit {s orig} {
	seshw::set_text $s [$orig.s.s get 0.0 end]
	seshw::autosize $s
	destroy $orig
    }

    proc insertWriting {this {txt ""}} {
	insertHieroglyphicTag $this WRI $txt
    }
    
    proc insertHiero {this {txt ""}} {
	insertHieroglyphicTag $this HIE $txt
    }

    proc insertDeterminative {this {txt ""}} {
	insertHieroglyphicTag $this DET $txt
    }
    
    
    # The cursor is "in" a tag iff
    # a) both sides of the cursor have the tag
    # b) the left side is NOT "end_TAG" or the right side is
    #    not "begin_TAG", in which case we are between two area marked with the tag

    # getTagsAtPos --
    # 	get the list of tag which can be written into at a given position.

    proc getTagsAtPos {this {pos insert}} {
	set path [getTextPath $this]
	
	set position [$path index $pos]
	if {[$path compare 1.0 == $position]} {
	    return {}
	}
	# tags after insert point 
	set l2 [$path tag names $position]
	# tags before insert point 
	set l1 [$path tag names "$position - 1 chars"]
	# these lists will be short, so we compute a simple intersection list.
	set l {}
	foreach i  $l1 {
	    # get tags which belongs to the two lists
	    if {[lsearch -exact $l2 $i] != -1} {
		# except when we have a closing tag before the point
		# or a begin tag after it.
		if {
		    [lsearch -exact $l1 end_$i] == -1 ||
		    [lsearch -exact $l2 begin_$i] == -1
		} {
		    lappend l $i
		}
	    }
	}
	# suppress the selection tag from the list.
	set spos [lsearch  -exact $l sel]
	if {$spos != -1} {
	    lreplace l $spos $spos
	}
	return $l
    }

    proc selectTag {this} {
	set path [getTextPath $this]
	# First, we find the previous begin tag for the current tag
	# next, we find the end tag. And voila, we have won
	set tag [getUpperTag $this]
	set txt [$path tag prevrange begin_$tag  insert]
	#regexp {$([^_]+)_} $tag tagname
	if { $txt == "" } {
	    return 0
	}
	set begin [lindex $txt 0]
	set txt [$path tag nextrange end_$tag  insert]
	if { $txt == "" } {
	    return 0
	}
	set end [lindex $txt 1]
	if [$path compare $end < insert] {
	    return 0
	}
	$path tag remove sel 0.0 end
	$path tag add sel $begin $end
	$path tag lower sel
    }
    
    ################################################################################
    #
    # Special bindings for Text widget containing hierarchic tags.
    # The idea is that text under these tags are 
    # a) not easy to delete 
    # b) impossible to edit. In fact, they are usually \n

    bind HierarchicText <Control-Tab> {
	focus [tk_focusNext %W]
    }
    
    bind HierarchicText <Control-Shift-Tab> {
	focus [tk_focusPrev %W]
    }
    
    bind HierarchicText <Delete> {DicoTk::callWindowCommand DicoTk::textdelete %W; break}
    
    bind HierarchicText <BackSpace> {DicoTk::callWindowCommand DicoTk::textbackspace %W; break}
    
    bind HierarchicText <KeyPress> {DicoTk::callWindowCommand DicoTk::keyinsert %W %A; break}
    
    bind HierarchicText <Return> {DicoTk::callWindowCommand DicoTk::keyinsert %W "\n"; break}
    
    bind HierarchicText <F10> {expr 1}
    
    bind HierarchicText <Button-3> {
	DicoTk::callWindowCommand DicoTk::buildTagSelectionMenu %W %X %Y
    }
    
    bind HierarchicText <Menu> {
	DicoTk::callWindowCommand DicoTk::buildTagSelectionMenu %W %X %Y
    }
    
    bind HierarchicText <Key-F2> {
	DicoTk::callWindowCommand DicoTk::buildTagSelectionMenu %W %X %Y
    }
    
    bind HierarchicText <Button-1> {
	focus %W
	%W mark set insert @%x,%y
	%W see insert
	DicoTk::callWindowCommand DicoTk::getUpperTag %W
    }
    

    bind HierarchicText <Left> {DicoTk::callWindowCommand DicoTk::textleft %W; break}
    bind HierarchicText <Right> {DicoTk::callWindowCommand DicoTk::textright %W; break}
    bind HierarchicText <Up> {DicoTk::callWindowCommand DicoTk::textup %W; break}
    bind HierarchicText <Down> {DicoTk::callWindowCommand DicoTk::textdown %W; break}
    bind HierarchicText <<Paste>> {DicoTk::callWindowCommand DicoTk::winpastetext %W; break}
}
