
#
# Contour Object :
#
# A Contour is a list of control points and segments
#
# Architecture :
#
# Un Contour comprend des points et des segments. Sa 
# destruction entraine la destruction de tous ceux-ci.
#
# A Contour contains points and segments. Destroying the contour
# destroys the segments and the points. HOWEVER, point destruction
# is something done by the segments, not the contour.
#
# Points and segments are linked. One must ATTACH points to segments, and free 
# them when they are not used.
#
# Note from a wiser(?) me : I was still young and foolish when designing the 
# font editor. Control is much too scattered among the various classes to be understandable.
# A stricter version of the editor should be made, with much stupider classes for 
# the model.

class Contour {

    proc Contour {this window {sens T}} GraphElt {$window} {
	set ($this,type) CMP
	set ($this,sens) $sens
	set ($this,selected) 0
	fixColor $this
	set ($this,segments) {}
	lappend CharEdit::($window,Contours) $this
    }

    proc ~Contour {this} {
	foreach s [PRIVgetSegments $this] {
	    delete $s
	}
	CharEdit::auxdeleteContour $GraphElt::($this,window) $this 
    }
    
    # Tag name used for main lines. Used 
    # for presentation-related stuff
    proc tagname {this} {
	return "tag$this"
    }

    # I put a "virtual" here. Can't remember if it's needed.
    virtual proc expandPoints {this} {
	set res {}
	return $res
    }

    proc PRIVgetSegments {this} {
	return $Contour::($this,segments)
    }

    # Get all control points for this contour
    proc PRIVgetControlPoints {this} {
	set ctrls [list]
	foreach s [PRIVgetSegments $this] {
	    eval lappend ctrls [lrange [Segment::getControlPoints $s] 1 end]
	}
	puts "controls : $ctrls"
	return $ctrls
    }

    #
    # Modify the contour according to the affine transformation transfo.

    proc appliTransformation {this transfo} {
	foreach p [PRIVgetControlPoints $this] {
	    AffineTransformation::apply $transfo $p
	}
    }

    proc createDrawing {this} {
	return [list]
    }

    proc update {this} {
	GraphElt::_update $this; # call parent class procedure
    }
    
    proc attachSegment {this seg} {
	lappend ($this,segments) $seg
    }

    proc detachSegment {this seg} {
	listreplace Contour::($this,segments) $seg
    }

    proc replaceSegment {this s newsegs} {  
	listreplace Contour::($this,segments) $s $newsegs
    }
    
    proc toggleSelect {this} {
	if $Contour::($this,selected) {
	    unselect $this
	} else {
	    select $this
	}
    }

    proc select {this} {
	set w $GraphElt::($this,window)
	set s $CharEdit::($w,selection)
	set ($this,selected) 1
	fixColor $this
	Selection::add $s $this
    }

    proc unselect {this} {
	set w $GraphElt::($this,window)
	set ($this,selected) 0
	fixColor $this
	Selection::suppress $CharEdit::($w,selection) $this
    }
    
    proc asText {this} {
	set res [list $Contour::($this,type) $Contour::($this,sens)]
	set fseg [lindex $Contour::($this,segments) 0]
	set res [concat $res [Point::asText [Segment::firstPoint $fseg]]]
	foreach s $Contour::($this,segments) {
	    lappend res [Segment::asText $s]
	}
	return $res
    }

    proc deletePointFrom {this point segments} {
	# If the point is an external control :
	# 0. between more than 3 segments : ????

	if {[llength $segments] != 2} return
    
	set s1 [lindex $segments 0]
	
	set s2 [lindex $segments 1]
	
	if {$s1 == $s2} return
    	
	set pos1 [lsearch -exact $Contour::($this,segments) $s1]
	set pos2 [lsearch -exact $Contour::($this,segments) $s2]
	if {$pos1 == -1 || $pos2 == -1} {error "heeeeeeeeeellllpppppp"}
	# Three real cases are possible :
	# the points follow one another
	# or one is the first point
	if {$pos1 == $pos2 + 1 || ($pos1 == 0 && $pos2 != 1)} {
	    set tmp $s1
	    set s1 $s2
	    set s2 $tmp
	}
	
	# (if none of these cases applied, pos2 is after pos1)
	
	set t 0
	
	set class1 [classof $s1]
	set class2 [classof $s2]
	
	if {[string compare $class1 "::LineSegment"] == 0} {incr t}
	if {[string compare $class2 "::LineSegment"] == 0} {incr t}
	
	# 0. between two splines : keep the external controls.
	# 1. between a line and a spline : create a spline
	# 2. between two lines : delete the point and create a new line.	

	set p1 [Segment::firstPoint $s1]
	set p2 [Segment::lastPoint $s2]
	
	switch -exact $t {
	    0 {
		set c1 [Spline::firstControl $s1]
		set c2 [Spline::secondControl $s2]
		Point::double $c1 $p1
		Point::double $c2 $p2
		set seg [new Spline $this $p1 $c1 $c2 $p2]
	    }
	    1 {
		if {[string compare $class1 "::LineSegment"] == 0} {
		    set c1 [Segment::lastPoint $s1]
		    set c2 [Spline::secondControl $s2]
		} else {
		    set c1 [Spline::firstControl $s1]
		    set c2 [Segment::firstPoint $s2]
		}
		#Point::double $c1 $p1
		#Point::double $c2 $p2
		set seg [new Spline $this $p1 $c1 $c2 $p2]
	    }
	    2 {
		set seg [new LineSegment $this $p1 $p2]
	    }
	}
	
	detachSegment $this $s2
	replaceSegment $this $s1 $seg
	delete $s1
	delete $s2
    }

    proc getFill {this} {
	return $($this,sens)
    }

    #	s is either "N" or "P".
    proc setFill {this s} {
	set ($this,sens) $s
	fixColor $this
    }

    proc fixColor {this} {
	if {[isSelected $this]} {
	    set ($this,color) yellow
	} else {
	    if  {[getFill $this] != "N"} {
		set ($this,color) black
	    } else {
		set ($this,color) darkgray
	    }
	}
	set w $GraphElt::($this,window)
	set p [Component::getPath $w]
	$p itemconfig [tagname $this] -fill $($this,color)
    }

    #	Returns : the *current* color of the contour.
    proc getColor {this} {
	return $($this,color)
    }

    proc isSelected {this} {
	return $($this,selected)
    }
}

