# hieroglyphs.tcl --
#
#       Common data about hieroglyphic signs
# 
#
# Remark : this file correspond to the time when signs where hard-coded
# now that they are defined in signedef, we should
# probably have a slightly different system.
#
# 
######################################################################
#
# Notes :
#
# Sign variant system :
# -------------------
#
# We shall use here a system with two levels of variants :
#
# a) sure variants : like Y1v for Y1, D40 for A24, Y2 for Y1 &.
#    These are variants which might mean something (for example, 
#    indicate a  hieratic original or date the text),
#    but don't lead to confusion. We don't use these variants 
#    for indexing the texts, however, because one might want 
#    to distinguish between them.
#
# b) synsets : the idea  is taken from a lexicology program called wordnet.
# a synset is a set of "synonymous signs". The synonymy relation
# is NOT transitive, however. A sign might belong to different synsets.
# 
# For example, there are a number of synset which contain D36 (the arm).
# One contains D46 (the hand), and another contains D40, but this does not make
# D46 and D40 synonymous signs. Typically, this relation can not be used in a
# full automatic way. One should be allowed to chose either a sign or a synset.
#
# Word looking system :
# ---------------------
#
# This system takes as its argument a list of signs. It can then either
#
# a) build a regular expression usable to look for the word.
# b) send back a list of equivalent writings
#
# The argument can contain :
#
# - signs : exact sign matching (gardiner code)
# - list of signs : matches one signe in the list
# - a "?" can precede the sign or list, in which case it's optionnal.
#
# e.g. : F40 ? G1 ? {G43 Z7} Y1
# will correspond to "Aw-Y1" or to "Aw-W-Y1", etc.
#
# implementation note :
#   The regular expression should be able to span lines.
#   between signs, we can have either a space, or "\n[^\t]*\t"
#   (the line number)

package provide hieroglyphs 0.0
package require glyphDAL


namespace eval hieroglyphs {
    variable bitmaps ;			# The loaded signs bitmap
    variable VAR
    variable variants;			# table of base for for signs which are variants
    variable synset;
    variable normcode;			# table of normal values for signs.
    
    # phonetic values and codes to base code.
    array set variants {}
    array set normcode {}
    
    set synset(n) 0;			# max synset number


    # List of signs in a section (A, B, etc...)

    proc getAllSignsInSection {letter} {
	set l [glyphDAL::getSignsCodesMatching "$letter\[0-9\]*"]
	return [lsort -command glyphCode::compareCodes $l]
    }

    # get all known values for a sign
    proc getAllSignValues {sign} {
	return [glyphDAL::getValuesForCode $sign]
    }

    # from a value (phonetic code or Gardiner code), build a list of 
    # possible official codes

    proc codeToSigns {codeOrValue} {
	#puts "getting $codeOrValue"
	if {[glyphCode::isGardiner $codeOrValue]} {
	    return [list $codeOrValue]
	} else {
	    # get all signs for whom "code" is a possible phonetic *value*
	    set result [glyphDAL::getCodesForValue $codeOrValue]
	    #puts "found signs $result"
	    
	    # if the code is an official phonetic code,
	    # then suppress sign from the list 
	    # in order to put it at the front position.

	    if {[glyphDAL::isOfficialPhoneticCode $codeOrValue]} {
		# suppress the code from the list :
		set indexcode [lsearch -exact $result $codeOrValue]
		if { $indexcode != -1} {
		    set result [lreplace $result $indexcode $indexcode]
		}
		# put it in the first position
		set result [concat $codeOrValue $result]
	    }
	    #puts "candidate list is $result"
	    return $result
	}
    }

    
    # Normalize a text (a list of signs).
    # 
    proc normalizeText {txt {minimal 1}} {
	set result ""
	if $minimal {
	    set norm normalize_sign
	} else {
	    set norm low_level_norm
	}
	foreach i $txt {
	    lappend result [$norm $i]
	}
	return $result
    }
    
    # declares sign code1 as a variant of code2
    proc defineVariant {code1 code2} {
	variable variants
	set variants($code1) $code2
	lappend variants(varset,$code1) $code2
	lappend variants(varset,$code2) $code1
    }
    
    # get all variants of a sign.
    proc getAllVariantsOf {code} {
	set l [glyphDAL::getVariants $code 5]
	puts $l
	return [concat $code $l]
    }
    
    # define set of "equivalent" signs.
    # the link of signs in a synset is weaker than 
    # the one between variants.
    proc define_synset {signs} {
	variable synset
	set l [normalizeText $signs 0]
	set n [incr synset(n)]
	set synset($n,signs) $l
	foreach s $l {
	    lappend synset($s,nums) $n
	}
    }
    
    # Get all signs that belong to a synset which contains code
    
    proc get_all_syns {code} {
	variable synset
	if [info exists synset($code,nums)] {
	    foreach i $synset($code,nums) {
		foreach s $synset($i,signs) {
		    lappend result $s
		}
	    }
	    return $result
	} else {
	    return $code
	}
    }
    
    # get a list of signs and sign group and generate all
    # possibilities
    # CURRENTLY NOT USED 

    proc expand_list {l} {
	set result {{}}
	if [llength $l] {
	    set opt 0
	    set h [lindex $l 0]
	    set l [lrange $l 1 end]
	    if { [string compare $h "?"] == 0} {
		set opt 1
		set h [lindex $l 0]
		set l [lrange $l 1 end]
	    }
	    
	    set tails [expand_list $l]
	    
	    if $opt {
		# as the head element is optional, the result
		# is initialized as :
		set result $tails
	    } else {
		set result {}
	    }
	    
	    foreach x $h {
		foreach e $tails {
		    lappend result [concat $x $e]
		}
	    }
	}
	return $result
    }
    
    # get a list of signs and sign group and create the corresponding regular expression.
    # 
    # Should probably stay in the search.tcl code.
    #
    # The expression is a bit cumbersome. It might be improved in two ways :
    # 
    # a) non optional lone signs don't need parenthesis at all
    # b) optional lone signs need only one set of parenthesis
    #
    # The  result can be used in various regexps
    #
    # e.g. transform a text into a list of 
    # the lines which contain the signs :
    #
    # set res [list_to_regexp ...]
    # regsub -all "(\[^\t\n\]+\t)\[^\n\]*$res\[^\n\]*|\[^\n\]*\n" $txt {\1}
    
    proc list_to_regexp {l} {
	#set space "( |(\n|\r|\r\n)\[^\t\]*\t)"
	set space "( | \n\[^\t\]*\t)"
	set r ""
	set opt 0
	foreach e $l {
	    if {[string compare $e "?"] == 0} {
		set opt 1
	    } else {
		append r "("
		append r "("
		append r [lindex $e 0]
		foreach x [lrange $e 1 end] {
		    append r "|"
		    append r $x
		}
		append r ")"
		append r " )"
		if $opt {
		    append r ?
		}
		set opt 0
	    }
	}
	regsub " (\[^ \]*)$" $r {\1} r
	regsub -all " " $r $space r
	return $r
    }

    # normalize a sign. 
    #
    # This procedure can be useful (for instance for exporting to other MdC softwares)
    # but it's somehow buggy as far as normalization is concerned.
    # The only way to go is to normalize to User Glyph Code, for they don't change.
    #
    # The rules are :
    #	official "gardiner" signs have their gardiner code.
    #	other signs have their gardiner codes.
    # 

    proc low_level_norm {code} {
	variable normcode
	# First, try the cache
	if [info exists normcode($code)] {
	    return $normcode($code)
	}

	set result $code
	switch -exact -- [glyphCode::getCodeType $code] {
	    N -
	    P {
		set r [glyphDAL::getSignCodeForPhoneticCode $code]
		if {[string length $r] != 0} {
		    set result $r
		}
	    }
	    G {
	    }
	    U {
		# look for Gardiner code. If none is found, use the user glyph code.
		set allcodes [glyphDAL::getEquivalentCodes $code]
		foreach c $allcodes {
		    if {[string equal "G" [glyphCode::getCodeType $c]]} {
			set result $c
		    }
		}
	    }
	    O {
		set gardinerCode ""
		set userGlyphCode ""
		set allcodes [glyphDAL::getEquivalentCodes $code]
		foreach c $allcodes {
		    switch -- [glyphCode::getCodeType $c] {
			G {set gardinerCode $c}
			U {set userGlyphCode $c}
		    }
		}
		if {[string length $gardinerCode] != 0} {
		    set result $gardinerCode
		} elseif {[string length $userGlyphCode] != 0} {
		    set result $userGlyphCode
		}
	    }
	}
	# puts "result for $code is $result"
	# saves the result in the cache
	set normcode($code) $result
	return $result
    }
    

    # return a *list* of possible equivalent for a sign.
    proc normalize_sign {code} {
	variable variants
	# a) first, we get Gardiner code.
	# 
	set code [low_level_norm $code]
	if [info exists variants($code)] {
	    return $variants($code)
	} else {
	    return $code
	}
    }
    
    proc getBitmapForSign {code} {	
	variable bitmaps
	if {[info exists bitmaps($code)]} {
	    return $bitmaps($code)
	} else {
	    # Load the bitmap definition from the base
	    set bm [glyphDAL::loadBitmapFromDB $code]
	    
	    # if it fails, build the bitmap definition
	    if {[string length $bm] == 0} {
		buildBitmapForSign $code
		set bm [glyphDAL::loadBitmapFromDB $code]
	    }
	    # Now, bm contains the bitmap.
	    # create an image : 
	    set bitmaps($code) [image create bitmap -data $bm]
	    return $bitmaps($code)
	}
    }

    # Private procedure
    proc buildBitmapForSign {code} {
	variable VAR
	if {! [winfo exists .signDummy]} {
	    set VAR(signDrawer) [seshw::create_passive .signDummy -htline 40]
	}
	seshw::set_text $VAR(signDrawer) $code
	seshw::autosize $VAR(signDrawer)
	set b [$VAR(signDrawer) toxbm 1.0 end]
	glyphDAL::saveBitmapInDB $code $b
    }

    # Callback used when a sign is unknown : loads it from the database.
    # Note that this code should probably be split in two :
    # * The "political part" should be here
    # * The actual database part should be in glyphDAL.

    # Callback used when a sign is unknown : loads it from the database.
    # Note that this code should probably be split in two :
    # * The "political part" should be in "hieroglyphs"
    # * The actual database part should be here.
    # to facilitate later migration to this, 
    # the actual callback is hieroglyphs::displayNotFound
    # 
    
    proc displayNotFound {CODE} {
	if {$CODE eq "." || $CODE eq ".."} return
	# Only create actual glyph for User glyph codes
	set codetype [glyphCode::getCodeType $CODE]
	switch -- $codetype {
	    U {	
		glyphDAL::applyToGlyphByUserCode $CODE tksesh_defchar
	    } 
	    N -
	    P  {
		glyphDAL::applyToGlyphByPhoneticCode $CODE hieroglyphs::PRIVregisterGlyphCode 1
	    }
	    O -
	    G {
		variable tempUCode
		set tempUCode ""
		glyphDAL::applyToGlyphByGardinerCode $CODE hieroglyphs::PRIVregisterGardinerCode 1
		# if an official code was found, tempUCode is not null :
		if {$tempUCode ne ""} {
		    tksesh_samesign $CODE $tempUCode
		} else {
		    # if no official code is available, go for non official ones.
		    if {! 
			[glyphDAL::applyToGlyphByGardinerCode \
			     $CODE hieroglyphs::PRIVregisterGlyphCode 0]} {
			glyphCode::addToUnknown $CODE
		    }
		}
	    }
	}
    }


    # called when registering a sign. Doesn't do the actual registering,
    # 
    proc PRIVregisterGardinerCode {gCode uid mid id} {
	variable tempUCode
	set tempUCode [glyphCode::buildCode $uid $mid $id]
    }

    proc PRIVregisterGlyphCode {gCode uid mid id} {
	tksesh_samesign $gCode [glyphCode::buildCode $uid $mid $id]
    }

}
