package provide tclSQLDB 1.0
package require sqlite

#
# the following line allows to use "observable". 
# the Observable class should obviously move.

package require myStoopWidgets

# Abstract ? class for all datasources

# Example of use : see DataSourceExamples.tcl


# DataSource uses the "Observable" Design Pattern.
# Messages sent by a DataSource to one of its observers :

# "ObservableDeleted"
# "IndexChanged"
# "dataChanged"
# "rowDeleted"
# "rowChanged"
# "rowCreated"
# "stateChanged"

# Column declaration : 
# a : autoincrement/automatic field.
# t : text field
# i : number field

## TO DO : add a "modified" flag to indicate that the data has been modified
#  (useful for "cancel" operation and the like)

#
# DataSources have a "state" :
# browse : moving through the available data. If there is data, only real rows will be 
# 	   available.
#
# edit : the current row is being edited. Editing can be either validated or canceled.
#
# createRow : in "createRow" mode, a new entry is being created.
#	moving to the "createRow" state moves the pointer to a special entry,
#	the "insert" entry.
# 
# deleteRow : in "deleteRow" mode. The current entry is marked for deletion.

class DataSource {
    
    ######################################################################
    #
    # Constructors and the like
    #

    proc DataSource {this db sql} Observable {} {
	set ($this,state) browse
	set ($this,db) $db
	set ($this,sql) $sql
	set ($this,index) 0
	set ($this,name) "anonymous"
	set ($this,filter) ""
	set ($this,bindings) {}; # see bindFields
	set ($this,modified) 0 ;# was the current row modified ?
	set ($this,validityTest) "" ; # a function called to test if this data is valid
	;			      # for insertion
	set ($this,editableTest) "" ; # a function called to test if data can 
	;                             #be edited for this row. Note that this is different 
	;                             #from the notion of editable *fields*
	setPrepareValidate $this "nop"; # Callback function called just before validating
	;                            # an entry and saving it in the database.
	;                            # in tksesh, used for timestamps and the like.
	# 
	set ($this,cacheLoaded) 0
	set ($this,firstInCache) 0
	set ($this,cacheSize) 100
	#
	set ($this,nbrRows) 0
	#
	setDeletableTest $this ""
	setUpdateRequest $this ""
	setDeleteRequest $this ""
	setCreateRequest $this ""
	setRefreshInsertEntryRequest $this "" ;# This request is used to update the insert entry.
	declareColumns $this {}
	declareExternal $this {}
	
	setResultKeptInMemory $this 1
    }

    proc ~DataSource {this} {
	Observable::notify $this "ObservableDeleted"
    }

    #
    # Build a Datasource for a table.
    # 	table : table name
    # 	idx : the list of index fields
    #   cols : declaration of all columns

    proc buildTableDataSource {db table idx cols} {
	set ds [new DataSource $db "select * from $table"]
	setName $ds $table
	declareColumns $ds $cols
	# declareExternal $ds $ext

	# build a list of all fields which don't belong to the key :
	foreach c [getColumnNames $ds] {
	    set tmp($c) 1
	}
	foreach c $idx {
	    unset tmp($c)
	}

	set notInKey [array names tmp]
	#
	
	# The part of the requests use to select on primary key :
	set selectKey ""
	if {[llength $idx] > 0} {
	    append selectKey "\n where \n"
	    set col [lindex $idx 0]
	    append selectKey "$col= :$col:"
	    foreach col [lrange $idx 1 end] {
		append selectKey " and $col= :$col:\n"
	    }
	}
	
	
	# Update request :
	if {[llength $notInKey] > 0} {
	    # Build update request 
	    # (doesn't mean anything if all fields belong to the key)
	    set request "update $table set "
	    set c [lindex $notInKey 0]
	    append request "$c= :$c:"
	    foreach col [lrange $notInKey 1 end ] {
		append request ", $col= :$col:"
	    }
	    append request $selectKey
	    setUpdateRequest $ds $request
	}
	
	# delete request :
	setDeleteRequest $ds "delete from $table $selectKey"
	
	# insert request :
	set createRequest "insert into $table ("
	set isFirst 1
	
	foreach col [getColumnNames $ds] {
	    if {! $isFirst } {
		append createRequest ", " 
	    }
	    append createRequest "$col"
	    set isFirst 0
	}
	append createRequest ") values ("
	set isFirst 1
	foreach col [getColumnNames $ds] {
	    if {! $isFirst } {
		append createRequest ", " 
	    }
	    if {[getColumnType $ds $col] eq "a"} {
		append createRequest "NULL"
	    } else {
		append createRequest ":$col:"
	    }
	    set isFirst 0
	}
	append createRequest ")"
	
	setCreateRequest $ds $createRequest
	return $ds
    }
			       
    proc declareColumns {this colDecl} {
	set ($this,columnNames) [PRIVAnalyseDeclaration $this $colDecl]
    }

    
    proc getColumnType {this col} {
	return $($this,type,$col)
    }

    proc declareExternal {this colDecl} {
	set ($this,externalFields) [PRIVAnalyseDeclaration $this $colDecl]
	PRIVSetNull $this $colDecl
    }

    proc getColumnNames {this} {
	return $($this,columnNames)
    }

    proc getExternalNames {this} {
	return $($this,externalFields)
    }

    proc getAllColumnNames {this} {
	return [concat [getColumnNames $this] [getExternalNames $this]]
    }

    # Sets a datasource name
    proc setName {this name} {
	set ($this,name) $name
    }

    # If result is kept in memory, the whole 
    # result for the select will be kept, which is probably the
    # reasonable thing to do in most cases for our databases.
    #
    # If large results are expected, set the flag to 0.
    # in this case, only N entries will be kept (N = ?)

    proc setResultKeptInMemory {this flag} {
	set ($this,resultKeptInMemory) $flag
    }

    proc isResultKeptInMemory {this} {
	return $($this,resultKeptInMemory)
    }
    
    # Can be used to modify an already created datasource,
    # in particular when made with buildTableDataSource
    
    proc setSelectRequest {this sql} {
	set ($this,sql) $sql
    }
    
    proc setUpdateRequest {this sql} {
	set ($this,updateRequest) $sql
    }

    proc setDeleteRequest {this sql} {
	set ($this,deleteRequest) $sql
    }

    proc setCreateRequest {this sql} {
	set ($this,createRequest) $sql
    }

    # This request, if defined, is called after an insert is validated.
    # It's used if the insertion process involves modifying the entry,
    # for instance with autoincrement fields.
    #
    # Currently not used !
    
    proc setRefreshInsertEntryRequest {this sql} {
	set ($this,refreshInsertEntryRequest) $sql
    }

    # Declare a column as editable.
    
    proc setEditable {this col {boolean 1}} {
	set ($this,isEditable,$col) $boolean
    }

    proc isEditable {this col} {
	return $($this,isEditable,$col)
    }
    
    # Declare a default value for a column.
    proc setDefault {this col value} {
	set ($this,default,$col) $value
    }
    
    # The filter is a set of *additionnal* conditions for a select, expressed in SQL.
    proc setFilter {this filter} {
	set ($this,filter) $filter
	refresh $this
	Observable::notify $this "dataChanged"
    }

    proc getFilter {this filter} {
	return $($this,filter) 
    }

    # setValidityTest --
    #	sets a function used to test if the current data can be inserted in the database.
    # args
    # 	func : a function which will be evaluated with this dataSource as argument.
    
    proc setValidityTest {this func} {
	set ($this,validityTest) $func
    }

    proc getValidityTest {this} {
	return $($this,validityTest)
    }

    # setDeletableTest --
    #	sets a function used to test if the current row can be deleted.

    proc setDeletableTest {this delete} {
	set ($this,deletableTest) $delete
    }

    proc getDeletableTest {this} {
	return $($this,deletableTest)
    }

    # setEditableTest --
    #	sets a function used to test if the current row can be edited.

    proc setEditableTest {this func} {
	set ($this,editableTest) $func
    }
    
    proc getEditableTest {this} {
	return $($this,editableTest)
    }
    
    # setPrepareValidate --
    # sets a function called before validating an entry. 
    # Can be used to sets a number of calculated values, etc...
    # 
    proc setPrepareValidate {this func} {
	set ($this,prepareValidate) $func
    }

    proc getPrepareValidate {this} {
	return $($this,prepareValidate)
    }

    ######################################################################
    #
    # getting information
    #

    # get the number of rows

    proc getSize {this} {
	return $($this,nbrRows)
    }

    # get the number of columns
    proc getNumberOfColumns {this} {
	return [llength [getColumnNames $this]]
    }

    # get the current index (between 0 and size - 1)
    # -1 means "no index".
    proc getIndex {this} {
	return $($this,index)
    }

    proc hasPrevious {this} {
	if {$($this,index) > 0 &&
	    $($this,index) < [getSize $this]} {
	    return 1
	} else {
	    return 0
	}
    }

    proc hasNext {this} {
	if {$($this,index) >= -1 &&
	    $($this,index) < [getSize $this] - 1} {
	    return 1
	} else {
	    return 0
	}
    }

    proc hasRow {this} {
	if {$($this,index) >= 0 &&
	    $($this,index) < [getSize $this]} {
	    return 1
	} else {
	    return 0
	}
    }
    
    proc getCol {this name} {
	if {[isNull $this $name]} {
	    error "$name is NULL"
	}
	return $($this,current,$name)
    }

    proc isNull {this name} {
	return $($this,current,$name,isNull)
    }

    # Gets the content of a column, using an index and the column name
    # usefull for listing the content of the whole datasource
    # without actually moving
    
    proc getColIdx {this idx col} {
	if {[isResultKeptInMemory $this]} {
	    return $($this,data,$idx,$col)
	} else {
	    PRIVrefreshCache $this $idx
	    return $($this,data,$idx,$col)
	}
    }

    proc PRIVclearCache {this} {
	set ($this,cacheLoaded) 0
	array unset DataSource:: "$this,data,*"
    }

    # test if the index is stored in cache, and reload cache if not.
    proc PRIVrefreshCache {this idx} {
	if {$($this,cacheLoaded) == 0 ||
	    $($this,firstInCache) > $idx || 
	    ($($this,firstInCache) + $($this,cacheSize)) <= $idx} {
	    set ($this,firstInCache) [expr {$idx - $($this,cacheSize)/2}]
	    if {$($this,firstInCache) < 0} {
		set ($this,firstInCache) 0
	    }
	    set sql "[PRIVPrepareRequest $this $($this,sql)\ $($this,filter)]"
	    set sql "select * from ($sql) \
                         limit $($this,cacheSize) offset $($this,firstInCache)"

	    PRIVclearCache $this
	    set i $($this,firstInCache)
	    [SQLDB::getDBCMD $($this,db)] eval $sql values {
		foreach c $values(*) {
		    set ($this,data,$i,$c) $values($c)
		}
		incr i
	    }
	    set ($this,cacheLoaded) 1
	}
    }


    
    #######################################################################
    # action system :
    #
    # The only legal way to change values in a DataSource is through actions.
    #
    # Possible actions are :
    #
    # * MOVING
    #	setIndex IDX : sets the index value to IDX
    #	next
    #	previous
    #	first
    #	last
    #
    # * EDITING :
    #	setCol NAME VALUE
    #	setNull NAME
    #
    # * CANCELING :
    #	cancel
    #
    # *VALIDATING EDITS/CREATION
    #	validate
    # 
    # * DELETING :
    #	deleteRow
    #
    # * creating a new entry :
    #	createRow
    #
    
    # Tests if a given action is currently possible for 
    # this DataSource.
    
    proc canDo {this action args} {
	set aclass $(aclass,$action)

	switch -- $aclass {
	    move {
		# if action is a move, first test if it's possible :
		switch -- $action {
		    setIndex {set idx [lindex $args 0]}
		    next {set idx [expr $($this,index) + 1]}
		    previous {set idx [expr $($this,index) - 1]}
		    first {set idx 0}
		    last {set idx [expr [getSize $this] - 1]}
		}
		if {$idx < 0 || $idx > [expr [getSize $this] - 1]} {
		    return 0
		}
	    }

	    edit {
		# Can't edit nothing.
		if {[isEmpty $this] && [getState $this] eq "browse"} {return 0}
		set func [getEditableTest $this]
		if {$func ne ""} {
		    if {! [eval $func $this] } {
			return 0
		    }
		}
		# For edits, the field must be editable.
		set col [lindex $args 0]
		if {! $($this,isEditable,$col)} {
		    return 0
		}
	    }
	    validate {
		set func [getValidityTest $this]
		if {$func ne ""} {
		    if {! [eval $func $this] } {
			return 0
		    }
		}
	    }
	    deleteRow {
		if {[isEmpty $this]} {return 0}
		set func [getDeletableTest $this]
		if {$func ne ""} {
		    if {! [eval $func $this] } {
			return 0
		    }
		}
	    }
	}
	# In all cases, the transition must not lead to the "no" state
	if {[nextState $this $aclass] eq "no"} {
	    return 0
	}
	# if we are here, the action is legal :
	return 1
    }

    proc doAction {this action args} {
	if {! [eval canDo $this $action $args]} {
	    return
	}
	[PRIVgetBeforeAction $this $(aclass,$action)] $this
	switch -- $action {
	    setIndex {
		PRIVFixIndex $this [lindex $args 0]
	    }
	    next {
		PRIVFixIndex $this [expr $($this,index) + 1]
	    }
	    previous {
		PRIVFixIndex $this [expr $($this,index) - 1]
	    }
	    first {
		PRIVFixIndex $this 0
	    }
	    last {
		PRIVFixIndex $this [expr [getSize $this] - 1]
	    }
	    setCol {
		PRIVSetCol $this [lindex $args 0] [lindex $args 1]
		set ($this,modified) 1
		Observable::notify $this "dataChanged"
	    }
	    setNull {
		set ($this,modified) 1
		PRIVSetNull $this [lindex $args 0] [lindex $args 1]
		Observable::notify $this "dataChanged"
	    }
	    cancel {
		PRIVLoadCurrent $this
	    }
	    validate {
		# actual validation is done when changing state.
	    }
	    deleteRow {
		PRIVDeleteRow $this
	    }
	    createRow {
	    }
	}
	PRIVChangeState $this $(aclass,$action)
    }
    
    
    #
    # Action(s) : moving
    #

    proc setIndex {this i} {
	doAction $this setIndex $i
    }

    proc next {this} {
	doAction $this next
    }

    proc previous {this} {
	doAction $this previous
    }
    
    proc last {this} {
	doAction $this last
    }

    proc first {this} {
	doAction $this first
    }

    ######################################################################
    #
    # edit a column value
    #

    proc setCol {this name val} {
	doAction $this setCol $name $val
    }
    
    # This action should probably be dealt with with doAction.
    proc setExternalCol {this name val} {
	if {! [hasValue $this $name $val]} {
	    PRIVSetCol $this $name $val
	    refresh $this
	    Observable::notify $this "dataChanged"
	}
    }
    
    proc setNull {this name {boolean 1}} {
	doAction $this setNull $name $boolean
    }

    # Tests if a datasource is empty (which kind of prevents editing and deleting)
    
    proc isEmpty {this} {
	return [expr {[getSize $this] == 0}]
    }
    # Tests if a given column has a given value
    proc hasValue {this name value} {
	if {[isNull $this $name]} {
	    return 0
	} else {
	    set v [getCol $this $name]
	    if {[getColType $this $name] eq "t"} {
		return [expr {$v eq $value}]
	    } else {
 		return [expr {$v == $value}]
	    }
	}
    }
    # Other actions :
    
    proc cancel {this} {
	doAction $this cancel
    }

    proc validation {this} {
	doAction $this validate
    }

    proc deleteRow {this} {
	doAction $this deleteRow
    }


    proc createRow {this} {
	doAction $this createRow
    }
    
    # bindings are a list of fields : {f1 fe1 f2 fe2 f3 fe3}
    # f1 is a normal field, fe1 is an external field whose value will be used
    # as default for f1 when creating a new value.
    #
    # This system is currently only used for new row creation.
    #
    proc bindFields {this bindings} {
	set ($this,bindings) $bindings
    }

    ######################################################################
    #
    # Private procedures
    #
    
    # Actually runs a request
    
    proc PRIVdoRequest {this type} {
	set sql $($this,${type}Request)
	if { $sql ne ""} {
	    set sql [PRIVPrepareRequest $this $sql]
	    SQLDB::update $($this,db) $sql
	} else {
	    error "don't know how to $type $($this,name)"
	}

    }

    # prepares a request for running. 
    # The fields for the current object are stored in the request as :
    # :fieldName: (the name of a field being made of letters, "_", and numerals.
    # PRIVPrepareRequest replaces these with the fields values

    proc PRIVPrepareRequest {this req} {
	set map {}
	foreach col [getAllColumnNames $this] {
	    set v [getColAsSQL $this $col]
	    lappend map ":$col:" $v
	}
	return [string map $map $req]
    }

    proc getColType {this col} {
	return $($this,type,$col)
    }

    proc getColAsSQL {this col} {
	if [isNull $this $col] {
	    return "NULL"
	} else {
	    switch -- $($this,type,$col) {
		a -
		i {set v $($this,current,$col)}
		t {set v "'$($this,current,$col)'"}
	    }
	    return $v
	}
    }
    
    # Analyses a declaration
    proc PRIVAnalyseDeclaration {this decl} {
	set columnNames {}
	foreach {col type} $decl {
	    lappend columnNames $col
	    set ($this,type,$col) $type
	    PRIVSetNull $this $col
	    if {$type eq "a"} {
		setEditable $this $col 0
	    } else {
		setEditable $this $col 1
	    }
	}
	return $columnNames
    }

    proc validate {this} {
	doAction $this validate
    }


    proc PRIVSetCol {this name val} {
	PRIVSetNull $this $name 0
	set ($this,current,$name) $val
    }
    
    proc PRIVSetNull {this name {boolean 1}} {
	set ($this,current,$name,isNull) $boolean
    }
    
    proc PRIVFixIndex {this idx} {
	set old $($this,index)
	set ($this,index) $idx
	if {$old != $($this,index)} {
	    PRIVLoadCurrent $this
	    Observable::notify $this "IndexChanged"
	}
    }
    
    
    # PRIVATE METHOD
    proc PRIVLoadCurrent {this} {
	if {[hasRow $this]} {
	    if { ! [isResultKeptInMemory $this]} {
		PRIVrefreshCache $this $($this,index)
	    }
	    foreach col [getColumnNames $this] {
		PRIVSetCol $this $col $($this,data,$($this,index),$col)
	    }
	} else {
	    foreach col [getColumnNames $this] {
		PRIVSetNull $this $col
	    }
	}
	set ($this,modified) 0
    }
    
    # Recompute the request
    proc refresh {this} {
	
	# Prepare the sql request.
	set sql [PRIVPrepareRequest $this "$($this,sql) $($this,filter)"]	    
	# Count the number of lines
	# This works for counting the number of lines for this source.
	# However, it's more or less as long as the whole computation (FALSE IN SOME CASES.) 
	set clock [clock clicks -milliseconds]
	set count [[SQLDB::getDBCMD $($this,db)] eval "select count(*) from ($sql)"]
	set c1 [clock clicks -milliseconds]

	if {[isResultKeptInMemory $this]} {
	    set i 0
	    [SQLDB::getDBCMD $($this,db)] eval $sql values {
		# We store everyting in an array :
		foreach col $values(*) {
		    set ($this,data,$i,$col) $values($col)
		}
		incr i
	    }
	    set ($this,nbrRows) $i
	    if {! [hasRow $this]} {
		if {$($this,index) >= [getSize $this]} {
		    set ($this,index) [expr [getSize $this] - 1]
		} else {
		    set ($this,index) 0
		}
	    }
	} else {
	    PRIVclearCache $this
	    set ($this,nbrRows) $count
	    # error "not implemented yet"
	}
	if {[hasRow $this]} {
	    PRIVLoadCurrent $this
	} else  {
	    PRIVSetDefaults $this
	}
	set c2 [clock clicks -milliseconds]
	PRIVNotifyChanges $this
    }
    
    proc setState {this state} {
	set ($this,state) $state
    }
    
    # Returns the state this DataSource is in
    # possible states are
    #
    # browse edit createRow deleteRow.
    proc getState {this} {
	return $($this,state)
    }

    # Action classes :
    set (aclass,setIndex) move
    set (aclass,next) move
    set (aclass,previous) move
    set (aclass,first) move
    set (aclass,last) move
    set (aclass,setCol) edit
    set (aclass,setNull) edit
    set (aclass,deleteRow) deleteRow
    set (aclass,cancel) cancel
    set (aclass,createRow) createRow
    set (aclass,validate) validate
    
    # Automaton for the datasource.
    # for each possible transition, data is a list of the form :
    # {target before after} 
    # where : 
    # 	target is the target state.
    #	before is a procedure to call before switching states
    #   after is a procedure to call after switching
    
    set (state,browse,edit) {edit nop PRIVNotifyChanges}
    set (state,browse,deleteRow) {browse nop nop}
    set (state,browse,move) {browse nop PRIVNotifyStateChange}
    set (state,browse,createRow) {createRow nop PRIVSetDefaults}
    set (state,browse,validate) {no nop nop}
    set (state,browse,cancel) {no nop nop}
    
    set (state,edit,edit) {edit nop nop}
    set (state,edit,deleteRow) {no nop nop}
    set (state,edit,move) {browse PRIVValidateEdit PRIVNotifyChanges}
    set (state,edit,createRow) {createRow PRIVValidateEdit PRIVSetDefaults}
    set (state,edit,validate) {browse PRIVValidateEdit PRIVNotifyChanges}
    set (state,edit,cancel) {browse nop PRIVNotifyChanges}

    set (state,createRow,edit) {createRow nop nop}
    set (state,createRow,deleteRow) {no nop nop}
    set (state,createRow,move) {no nop nop}
    set (state,createRow,createRow) {createRow PRIVValidateCreateRow PRIVSetDefaults}
    set (state,createRow,validate) {browse PRIVValidateCreateRow PRIVNotifyChanges}
    set (state,createRow,cancel) {browse nop PRIVNotifyChanges}

    set (state,no,edit) {no nop nop}
    set (state,no,deleteRow) {no nop nop}
    set (state,no,move) {no nop nop}
    set (state,no,createRow) {no nop nop}
    set (state,no,validate) {no nop nop}
    set (state,no,cancel) {no nop nop}

    
    # Compute next state for an action :
    proc nextState {this action} {
	return [lindex $(state,[getState $this],$action) 0]
    }

    # "stateChanged" means that the *possible* operations might have changed.
    # 
    proc PRIVNotifyStateChange {this} {
	Observable::notify $this "stateChanged"
    }
    
    proc PRIVNotifyChanges {this} {
	Observable::notify $this "dataChanged"
    }

    proc PRIVgetBeforeAction {this action} {
	return [lindex $(state,[getState $this],$action) 1]
    }

    proc PRIVgetAfterAction {this action} {
	return [lindex $(state,[getState $this],$action) 2]
    }
    
    proc PRIVChangeState {this action} {
	set st [getState $this]
	set dest [nextState $this $action]
	set command [PRIVgetAfterAction $this $action]
	setState $this $dest
	$command $this
    }

    proc PRIVValidateEdit {this} {
	if {[canDo $this validate]} {
	    eval [getPrepareValidate $this] $this
	    PRIVdoRequest $this update
	    #refresh $this
	    foreach col [getColumnNames $this] {
		catch {
		    set idx $($this,index)
		    set ($this,data,$idx,$col) [getCol $this $col]
		}
	    }
	    # Observable::notify $this "dataChanged"
	}
    }
    
    proc PRIVValidateCreateRow {this} {
	if {[canDo $this validate]} {
	    eval [getPrepareValidate $this] $this
	    PRIVdoRequest $this create
	    refresh $this
	    Observable::notify $this "rowCreated"
	}
    }

    proc PRIVDeleteRow {this} {
	PRIVdoRequest $this delete
	refresh $this
	Observable::notify $this "rowDeleted"
    }

    proc PRIVSetDefaults {this} {
	foreach col [getColumnNames $this] {
	    PRIVSetDefault $this $col
	}
	foreach {col ecol} $($this,bindings) {
	    PRIVSetCol $this $col [getCol $this $ecol]
	}
	Observable::notify $this "dataChanged"
    }

    proc PRIVSetDefault {this col} {
	if {[info exists DataSource::($this,default,$col)]} {
	    PRIVSetCol $this $col $($this,default,$col)
	} else {
	    PRIVSetNull $this $col
	}
    }
    
    # Do nothing
    proc nop {this} {
    }
}
