#=======================================================================#
# SCRIPT  : viewIcons.tcl                                               #
# PURPOSE : Display icons from icon library.                            #
# AUTHOR  : Adrian Davis (adrian@satisoft.com).                         #
#-----------------------------------------------------------------------#
# HISTORY : Mar02 1.00.00 - First release.                              #
#=======================================================================#

lappend auto_path [pwd]

#=======================================================================#
# Name      : initInfo                                                  #
# Purpose   : Initialise icon info lookup.                              #
# Parameters: Item     : Item to which "IconName" applies.              #
#           : IconName : Name of icon.                                  #
#=======================================================================#

proc initInfo {Item IconName} {
   global INFO

   set INFO($Item) $IconName

   bind $Item <Enter> {delayInfo %W}
   bind $Item <Leave> {cancelInfo}
}

#=======================================================================#
# Name      : delayInfo                                                 #
# Purpose   : Set delay for icon info.                                  #
# Parameters: Item : Item to which help applies.                        #
#=======================================================================#

proc delayInfo {Item} {
   global INFO

   cancelInfo
   set INFO(Delay) [after 200 [list showInfo $Item]]
}

#=======================================================================#
# Name      : cancelInfo                                                #
# Purpose   : Cancel icon info.                                         #
# Parameters: None.                                                     #
#=======================================================================#

proc cancelInfo {} {
   global INFO

   if {[info exists INFO(Delay)]} {
      after cancel $INFO(Delay)
      unset INFO(Delay)
   }
   wm withdraw .infoWindow
}

#=======================================================================#
# Name      : showInfo                                                  #
# Purpose   : Show icon info.                                           #
# Parameters: Item : Item to which info applies.                        #
#=======================================================================#

proc showInfo {Item} {
   global ICON INFO

   set IconName $INFO($Item)

   .infoWindow.name   configure -text [lindex $ICON($IconName) 0]
   .infoWindow.groups configure -text [lindex $ICON($IconName) 1]
   .infoWindow.type   configure -text [lindex $ICON($IconName) 2]
   .infoWindow.size   configure -text [lindex $ICON($IconName) 3]

   set InfoX [expr [winfo rootx $Item] + 10]
   set InfoY [expr [winfo rooty $Item] + [winfo height $Item]]
 
   wm geometry  .infoWindow +$InfoX+$InfoY
   wm deiconify .infoWindow

   raise .infoWindow

   unset INFO(Delay)
}

#=======================================================================#
# Name      : layoutResize                                              #
# Purpose   : Set scrolling region size.                                #
# Parameters: None.                                                     #
#=======================================================================#

proc layoutResize {} {

   set Box   [.display.icons bbox all]
   #set Width [expr {[winfo width .controls] - [winfo width .display.yscrollbar] - 8}]
   set Width [expr {[winfo width .controls] - [winfo width .display.yscrollbar] + 20}]

   .display.icons configure -width $Width -scrollregion $Box -xscrollincrement 0.1i -yscrollincrement 0.1i
}

#=======================================================================#
# Name      : selectIcons                                               #
# Purpose   : Display icons file selection dialog.                      #
# Parameters: None.                                                     #
#=======================================================================#

proc selectIcons {} {
   global LIBRARY

   set OldLibrary $LIBRARY

   set LIBRARY [tk_getOpenFile -initialdir . -initialfile tkIcons -title "Select Icon Library"]

   if {[string compare $LIBRARY ""] == 0} {
      set LIBRARY $OldLibrary
   }

   displayIcons
}

#=======================================================================#
# Name      : displayIcons                                              #
# Purpose   : Display icons from selected library.                      #
# Parameters: None.                                                     #
#=======================================================================#

proc displayIcons {} {
   global GROUPS ICON ICONS LIBRARY

   set Column 0
   set Row    0

   if {! [file exists $LIBRARY]} {
      tk_messageBox -icon warning -message "File does not exist" -title "viewIcons"
      return
   }


   set Cursor [. cget -cursor]
   . configure -cursor watch


   ::icons::icons delete $ICONS
   destroy .display.icons.layout.grid
   frame   .display.icons.layout.grid
   pack    .display.icons.layout.grid

   set ICONS [::icons::icons create -file $LIBRARY -group $GROUPS]

   foreach IconInfo [::icons::icons query -file $LIBRARY -group $GROUPS] {
      set IconName        [lindex $IconInfo 0]
      set ICON($IconName) $IconInfo

      label .display.icons.layout.grid.$Column:$Row -image ::icon::$IconName
      grid  .display.icons.layout.grid.$Column:$Row -column $Column -row $Row -padx 3 -pady 3
      grid  columnconfigure .display.icons.layout.grid.$Column:$Row $Column

      initInfo .display.icons.layout.grid.$Column:$Row $IconName

      if {$Column == 16} {
         set Column 0
         set Row [expr {$Row + 1}]
      } else {
         set Column [expr {$Column + 1}]
      }
   }  

   . configure -cursor $Cursor
}

#=======================================================================#
# Main code.                                                            #
#=======================================================================#

package require icons 1.0

#-----------------------------------------------------------------------#
# Initialise font.                                                      #
#-----------------------------------------------------------------------#

set PLATFORM $tcl_platform(platform)

if {[string compare $PLATFORM "unix"] == 0} {
   option add *font {Helvetica 10}
   option add *HighlightThickness 0
}

#-----------------------------------------------------------------------#
# Initialise Globals.                                                   #
#-----------------------------------------------------------------------#

label .dummy

set FONT     [.dummy cget -font]
set BOLDFONT "$FONT bold"
set GROUPS   "*"
set ICONS    {}
set LIBRARY  [file join [info library] tkIcons]

#-----------------------------------------------------------------------#
# Create main display window/frames.                                    #
#-----------------------------------------------------------------------#

frame .controls

label  .controls.libraryLabel -font $BOLDFONT -text "Library"
label  .controls.groupsLabel  -font $BOLDFONT -text "Groups"
entry  .controls.library      -width 50 -textvariable LIBRARY
entry  .controls.groups       -width 50 -textvariable GROUPS
button .controls.browse       -text "Browse" -command selectIcons
button .controls.view         -text "View"   -command displayIcons
frame  .controls.line -height 2 -borderwidth 1 -relief sunken

grid .controls.libraryLabel -row 0 -column 0 -padx 4
grid .controls.library      -row 0 -column 1
grid .controls.browse       -row 0 -column 2 -padx 4 -pady 2 -sticky ew
grid .controls.line         -row 1 -column 0 -pady 2 -columnspan 3 -sticky ew
grid .controls.groupsLabel  -row 2 -column 0 -padx 4
grid .controls.groups       -row 2 -column 1
grid .controls.view         -row 2 -column 2 -padx 4 -pady 2 ;#-sticky ew
pack .controls

bind .controls.library <Return> displayIcons
bind .controls.groups  <Return> displayIcons

frame     .display -borderwidth 2 -relief sunken
scrollbar .display.xscrollbar -command ".display.icons xview" -orient horizontal
scrollbar .display.yscrollbar -command ".display.icons yview" -orient vertical
canvas    .display.icons -xscrollcommand ".display.xscrollbar set" -yscrollcommand ".display.yscrollbar set"
pack      .display.yscrollbar -side right  -fill y
pack      .display.xscrollbar -side bottom -fill x
pack      .display.icons      -side left   -fill both -expand yes

frame     .display.icons.layout
pack      .display.icons.layout

.display.icons create window 0 0 -anchor nw -window .display.icons.layout

bind .display.icons.layout <Configure> "layoutResize"

pack .display -expand yes -fill both

wm title . "viewIcons"

layoutResize

wm resizable . 1 1

#-----------------------------------------------------------------------#
# Create "window" for icon info.                                        #
#-----------------------------------------------------------------------#

toplevel .infoWindow -background lightyellow -borderwidth 1 -relief solid

label .infoWindow.nameLabel   -background lightyellow -font $BOLDFONT -justify left -text "Name"
label .infoWindow.groupsLabel -background lightyellow -font $BOLDFONT -justify left -text "Groups"
label .infoWindow.typeLabel   -background lightyellow -font $BOLDFONT -justify left -text "Type"
label .infoWindow.sizeLabel   -background lightyellow -font $BOLDFONT -justify left -text "Size"
label .infoWindow.name        -background lightyellow -justify left
label .infoWindow.groups      -background lightyellow -justify left
label .infoWindow.type        -background lightyellow -justify left
label .infoWindow.size        -background lightyellow -justify left

grid  .infoWindow.nameLabel   -row 0 -column 0 -sticky w
grid  .infoWindow.name        -row 0 -column 1 -sticky w
grid  .infoWindow.groupsLabel -row 1 -column 0 -sticky w
grid  .infoWindow.groups      -row 1 -column 1 -sticky w
grid  .infoWindow.typeLabel   -row 2 -column 0 -sticky w
grid  .infoWindow.type        -row 2 -column 1 -sticky w
grid  .infoWindow.sizeLabel   -row 3 -column 0 -sticky w
grid  .infoWindow.size        -row 3 -column 1 -sticky w

wm overrideredirect .infoWindow 1
wm withdraw .infoWindow

#=======================================================================#
# End of script: viewIcons.tcl                                          #
#=======================================================================#
